#include "BiasedResidualsProcessor.h"

//C++
#include <iostream>
using namespace std;

//LCIO
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include "IntersectionCalculator.h"

// Aida
//#ifdef MARLIN_USE_AIDA // the whole processor does not make sense without aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
// endif

// Marlin 
#include <marlin/Global.h>

// GEAR
#include <gear/TPCParameters.h>

#include "TrackFitterFactory.h"
#include "TrackFitterBase.h"

using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

BiasedResidualsProcessor aBiasedResidualsProcessor;
    
BiasedResidualsProcessor::BiasedResidualsProcessor(): Processor("BiasedResidualsProcessor")
{
    // register steering parameters: name, description, class-variable, default value
    registerInputCollection(    LCIO::TRACK,
				"InputTracks" , 
				"Name of the input tracks collection"  ,
				_inputColName ,
				std::string("TPCSeedTracks") ) ;
    // modify processor description
    _description = "BiasedResidualsProcessor: Fill biased residuals of hits in tracks into AIDA histgram" ;

  registerProcessorParameter( "HistogramMinValueXY" , 
			      "Minimum value in xy histogram (default: -0.3 mm)"  ,
			      _histogramMinValueXY ,
			      float(-0.3) ) ;

  registerProcessorParameter( "HistogramMaxValueXY" , 
			      "Maximal value in xy histogram (default:  0.3 mm)"  ,
			      _histogramMaxValueXY ,
			      float(0.3) ) ;

  registerProcessorParameter( "HistogramMinValueZ" , 
			      "Minimum value in z histogram (default: -1.0 mm)"  ,
			      _histogramMinValueZ ,
			      float(-1.0) ) ;

  registerProcessorParameter( "HistogramMaxValueZ" , 
			      "Maximal value in z histogram (default:  1.0 mm)"  ,
			      _histogramMaxValueZ ,
			      float(1.0) ) ;

  registerProcessorParameter("ApplyCorrectionFactor" ,
			     "Multiply residual with sqrt{n / (n-dof)} (default: true)" ,
			     _applyCorrectionFactor ,
			     bool (true) ) ;

  registerProcessorParameter("DoFXY" ,
			     "Degrees of freedom in the XY plane (2 for straight line, 3 for helix) (default: 3)",
			     _dofXY,
			     int (3) ) ;
}

BiasedResidualsProcessor::~BiasedResidualsProcessor()
{
}

void BiasedResidualsProcessor::init()
{ 
	// usually a good idea to
	printParameters() ;

	// get histograms from the AIDAProcessor
	_residualsXYHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "residualsXYHisto", 
			     "residualsXYHisto", 
			     200, _histogramMinValueXY, _histogramMaxValueXY);
	_residualsZHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "residualsZHisto", 
			     "residualsZHisto", 
			     200, _histogramMinValueZ, _histogramMaxValueZ);

	// the plane extend is taken from GEAR
	const  std::vector< double > &planeExtent = Global::GEAR->getTPCParameters().getPlaneExtent();

	_residualsXY_vs_xr =   AIDAProcessor::histogramFactory(this)->
	    createHistogram2D( "residualsXY_vs_xr", 
			       "residualsXY_vs_xr", 
			       100, _histogramMinValueXY, _histogramMaxValueXY,
		               100, planeExtent[0], planeExtent[1]);

 	_residualsXY_vs_yphi =   AIDAProcessor::histogramFactory(this)->
	    createHistogram2D( "residualsXY_vs_yphi", 
			       "residualsXY_vs_yphi", 
			       100, _histogramMinValueXY, _histogramMaxValueXY,
		               100, planeExtent[2], planeExtent[3]);
   }

void BiasedResidualsProcessor::processRunHeader( EVENT::LCRunHeader* run )
{
}
    
void BiasedResidualsProcessor::processEvent( LCEvent * evt ) 
{ 

    // we need a LCCollectionVec to use iterators, as LCCollectio is not a std::vector
    LCCollectionVec* inputCollection=0;
    try{
	inputCollection = dynamic_cast<LCCollectionVec*>(evt->getCollection( _inputColName )) ;
    }
    catch(DataNotAvailableException &e)
    {
	// no input data
	// just continue ...
	return;
    }

    if( inputCollection != 0 )
    {
	int trackNr=0;

        LCParameters const & trackCollectionParameters = inputCollection->getParameters();

	for(LCCollectionVec::iterator it = inputCollection->begin() ; 
	    it < inputCollection->end(); it ++ )
	{
	    
	    streamlog_out(DEBUG4) <<"\t TrackNumber "<<trackNr++<<std::endl;

	    Track * track = dynamic_cast<Track *>(*it) ;

	    // first try to get a fitter, according to the fitter type defined in the track
	    unsigned char fitterType = TrackFitterBase::getFitterTypeFromTrack( track->getType() );

	    if (fitterType == 0) // no fitter has been set
	    {
	        fitterType = TrackFitterBase::FitterTypes::SIMPLEMINIMIZER;
	    }
	    
	    streamlog_out(DEBUG4) << "Using fitter type "<<static_cast<unsigned short>(fitterType)<< " to calculate residuals"<<std::endl;

	    TrackFitterBase *trackFitter = TrackFitterFactory::getTrackFitter( fitterType ,
									       & trackCollectionParameters);
	    
	    TrackerHitVec hitVector = track->getTrackerHits();
	    
//	    cout << "nr of hits " << hitVector.size()<< endl;

//	    int hitNr=0;

//	    cout << "\t\t hit\t x \t y \t z \t residual_x/r*phi" << endl;
	    for (TrackerHitVec::iterator hit = hitVector.begin();
		 hit <  hitVector.end(); hit ++)
	    {
		const double *position = (*hit)->getPosition();

		try
		{
		    EVENT::DoubleVec residuals = trackFitter->calculateResiduals(*hit, track);
		    _residualsXYHisto->fill( residuals[0] * (_applyCorrectionFactor?
							     sqrt(hitVector.size() / ( hitVector.size() - _dofXY) ):
							     1) );
		    _residualsZHisto->fill(  residuals[1]  * (_applyCorrectionFactor?
							      sqrt(hitVector.size() / (hitVector.size()-2) ):
							      1) );

		    _residualsXY_vs_xr->fill( residuals[0], position[0] );
		    _residualsXY_vs_yphi->fill( residuals[0], position[1] );

//		    cout << residuals[0] << "\t" << residuals[1]  << std::endl;
		}
		catch ( IntersectionCalculator::NoIntersectionException &e )
		{
		    cout << e.what() << endl;
		}
	
	    }
        }

    }
}


}// namespace 
