#ifndef DriftProcessor_h
#define DriftProcessor_h 1


//LCIO
#include <EVENT/SimTrackerHit.h>
#include <lcio.h>

//MARLIN
#include "marlin/Processor.h"

//C++
#include <string>

//GSL
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>



namespace marlintpc{


/* The following comment is tranfered to doxygen documentation
 * which begins with a second asterix, or in case of on line comment an extra third slash 
 */

/** Processor to drift primary electrons according to the
 *  characteristics of the choosen gas to pad plane. 
 *  Longitudinal and transverse diffusion are taken into
 *  account as well as magnetic fields.
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection of SimTrackerHits with charge values set as dEdx
 *
 *  <h4>Output</h4> 
 *  Collection containing SimTrackerHits with x,y-position on
 *  pad plane.
 *
 * @param InputCollectionName Name of input collection, SimTrackerHits
 * @param OutputCollectionName Name of output collection, SimTrackerHits
 * @param Transient If not 0 the output collection is set transient
 * @param Gas The gas in the TPC
 * @param DriftField Value of the drift field in V/cm
 *
 * @author Astrid Muennich 
 */

class DriftProcessor : public marlin::Processor {
  
 public:
  
  virtual Processor*  newProcessor() { return new DriftProcessor ; }
  
  /** C'tor
   */
  DriftProcessor() ;

  /** D'tor
   */
  virtual  ~DriftProcessor(){}  

  /** Called at the begin of the job before anything is read.
   * Use to initialize the processor, e.g. book histograms.
   */
  virtual void init() ;
  
  /** Called for every run.
   */
  virtual void processRunHeader( lcio::LCRunHeader* run ) ;
  
  /** Called for every event - the working horse.
   */
  virtual void processEvent( lcio::LCEvent * evt ) ; 


  /** Called after all events were processed for clean up
   */
  virtual void end();
  

  void InitializeDriftValues();

  /** Calculates transversal diffusion 
   */
  void CalculateTransverseDiffusion();

  /** Calculates coordinates of the drifted electrons. Longitudinal as 
   *  well as transvers diffusion are taken into account.
   */
  void MakeElectronDriftCoordinates();

  /** Gets electron coordinates before drifting
   */
  void GetElectronCoordinates(SimTrackerHit *);

  /** Calculates gas parameters from MAGBOLTZ parametrisations
   */
  double CalculateGasParameter(double a, double b ,double c, double d,
			       double e, double f, double g, double h,
			       double i, double j,double k);

  /** Internal function to take reduction of transversal diffusion in magnetic 
   *  fields into account
   */
  double CalculateDiffTransBParam(double a, double b, double c,double d, double e);

protected:

  //Input/output collection names
  std::string _inputCollectionName;///<Name of the input collection
  std::string _outputCollectionName;///<Name of the output collection
  std::string _tpcConditionsColName;///<Name of collection with TPCConditions
  std::string gas;///<Name of the gas used in the TPC
  const char* Gas;
  
  //For internal bookkeeping
  int _nRun ;
  int _nEvt ;
 
  float E;
  float _BField;

  int _transient;///<If not set to 0 output collection is set transient.
 
  gsl_rng *r;///<Random generator for landau and gaussian 
  // coordinates and variables for track
  /*! Coordinates of the electron. */
  double x_e,y_e,z_e,t_e; // e- coordinates
  // particle variables
  //int numberofelectrons;

  //TPC parameter
  double _TPCLength;///<Maximum drift length of the TPC
  double _TPCOuterRadius;///<Outer Radius of the TPC
  double _TPCInnerRadius;///<Inner Radius of the TPC
  double _distance;///<Drift distance of an electron

  //Coordinates of the drifted electron.
  float xdrift,ydrift,zdrift,tdrift; ///<Coordinates of drifted e-
  float vdrift;  ///<Drift velocity in  mm/us  
  float difflong; ///<Longitudinal diffusion in um/sqrt(cm)
  float difftrans; ///<Transversal diffusion in um/sqrt(cm)
  double _meanAmp;///<Mean amplification of an single electron.
};

}//end of namespace marlintpc

#endif 
