#ifdef USE_LCCD

#ifndef PHOTOELECTRICPROJECTIONFINDER_H
#define PHOTOELECTRICPROJECTIONFINDER_H 1

// Marlin
#include <marlin/Processor.h>

// LCIO
#include <lcio.h>

// gear
#include <gearimpl/GlobalPadIndex.h>

// stl
#include <map>
#include <set>
#include <string>
#include <vector>

namespace CLHEP {

	class Hep2Vector;
}

namespace EVENT {

	class TrackerPulse;
}

namespace marlintpc {

/** PhotoelectricProjectionFinder: groups the pulses produced by the photoelectric system into clusters. These clusters are then seperated into two groups: photodots and photolines.
 *    @author J. Abernathy (UVic)<br>
 */
class PhotoelectricProjectionFinder : public marlin::Processor {

public:

	PhotoelectricProjectionFinder() ;

	virtual void init() ;

	virtual Processor*  newProcessor() { return new PhotoelectricProjectionFinder(); }

	virtual void processRunHeader( lcio::LCRunHeader* run );

	virtual void processEvent( lcio::LCEvent * evt );

	virtual void check( lcio::LCEvent * evt );

	virtual void end();

private:

	/** the name of the input collection of TPCPulses found by a pulse finder
	*/
	std::string _input_tracker_pulses_collection_name;

	/** specifies the maximum distance (in mm) that two points can be separated before they are considered to be neighbours
	 */
	float _max_separation;

	/** specifies the minimum density that a point has to surpass to create a cluster. The units are the same as the units of the charge in the TrackerPulse
	 */
	float _density_threshold;

	/** specifies the name of the collection which stores the results of the projection finding
	 */
	std::string _projected_photogeometry_file_name;

	/** this file provides the locations of the photodots and photolines on the TPC cathode
	 */
	std::string _photogeometry_file_name;

	/** this contains the "weight" of each pad... how many times it has been seen in this run
	 */
	std::map< gear::GlobalPadIndex, int > _counts_by_pad;

	/** the total number of events which have been processed
	 */
	int	_total_event_count;

	/** the beginning (in ns) of the interval of the arrival time for photoelectric data
	 */
	float _photopulse_interval_start;

	/** the end (in ns) of the interval of the arrival time for photoelectric data
	 */
	float _photopulse_interval_end;

	/** this is fraction of events which a pad has to be in to be extracted for having photoelectric data
	 */
	float _event_fraction_threshold;

	/** this function takes a collection of tracker pulses and produces a set of points and a mapping of a weight to each point
	 */
	void create_points ( std::vector<CLHEP::Hep2Vector*> &points, std::map< CLHEP::Hep2Vector*, gear::GlobalPadIndex> &pad_by_point );

	/** this function takes a collection of points and produces a map of neighbours of each point within a provided distance of each other
	 */
	void find_neighbours ( const std::vector<CLHEP::Hep2Vector*> &points, const float max_seperation, std::map< CLHEP::Hep2Vector*,std::set<CLHEP::Hep2Vector*> > &neighbours_by_point );

	/** this function takes a collection of points, their neighbours and weights and produces clusters. The parameter density_threshold specifies the density which is required to produce a cluster. The density is calculated as the sum of the weights of a point and it's neighbours
	 */
	void calculate_point_clusters ( const std::vector< CLHEP::Hep2Vector* > &points, const std::map< CLHEP::Hep2Vector*, std::set< CLHEP::Hep2Vector* > > &neighbours_by_point, std::set< std::set< CLHEP::Hep2Vector* > > &point_clusters );

	/** this function takes the collections of points and sorts them into one of three categories: photodots, photolines or noise
	 */
	void classify_point_clusters( const std::set< std::set< CLHEP::Hep2Vector* > > &point_clusters, const std::map< CLHEP::Hep2Vector*, gear::GlobalPadIndex > pad_by_point, std::set< const std::set< CLHEP::Hep2Vector*>* > &photodots, std::set< const std::set< CLHEP::Hep2Vector* >* > &photolines );

	/** this function takes the sorted collections, converts them to LCIO TrackerHit classes and calculates the seed parameters for the fitting
	 */
	void store_results ( const std::set< const std::set< CLHEP::Hep2Vector* >* > &photodots, const std::set< const std::set< CLHEP::Hep2Vector* >* > &photolines, std::map< CLHEP::Hep2Vector*, gear::GlobalPadIndex > pad_by_point );

	/** this function produces some tuples for analysis
	 */
	void produce_tuples ( const std::set< const std::set< CLHEP::Hep2Vector* >* > &photodots, const std::set< const std::set< CLHEP::Hep2Vector* >* > &photolines );
};

}

#endif // PHOTOELECTRICPROJECTIONFINDER_H
#endif // USE_LCCD
