#ifndef ATTKALTRACK_H
#define ATTKALTRACK_H

#include <map>
#include <list>
#include "EVENT/TrackerHit.h"
#include "IMPL/TrackImpl.h"
#include "TPCHitMap.h"
#include "TPCHitDomain.h"
#include "gear/TPCParameters.h"
#include "TMatrixD.h"

using namespace EVENT;
using namespace std;
using namespace IMPL;

namespace marlintpc 
{

    	/** This is a base track class. It provides common methods for tracking.
     	 *  
     	 *  @author Bo Li, Tsinghua University
     	 *  @author Keisuke Fujii, KEK
     	 */

class AttKalTrack
{
public:
	AttKalTrack(const TPCParameters* tpcpara);

	AttKalTrack(const AttKalTrack& orig);

	virtual ~AttKalTrack();

	/** Create a track by copying from the original one
         */
	virtual AttKalTrack* create     (const AttKalTrack * orig=0)=0;

        /** Reset parameters and data of track
         */
	virtual void         reset      ()=0;

        /** Try to find a track and complete data operation
         */
	virtual TrackImpl*   findTrack  (ModuleRowHit hit);

        /** Search a proper hit to track
         */
	virtual void         searchHit  ();

        /** Search track starting from hit
         */
	virtual bool         searchTrack(ModuleRowHit hit)=0;

        /** Set hits container for track
         */
	inline void          setHitMap  (TPCHitMap *map){_tpcHitMap = map;}

        /** Remove hits stored in track object
         */
	virtual void         removeHits ();

        /** Add a new site to track and update track parameters
         */
	virtual bool      doAddAndFilter(ModuleRowHit hit)=0;

        /** Get which type this track is, helix or straight line.
         */
	virtual int       getType     ()=0;

        /** Get the current state vector the track
         */
  	virtual TMatrixD  getStateVec ()=0;

        /** Get the Chi2 increment at one step
         */
  	virtual double    getDeltaChi2()=0;

        /** Get hit number in the current track
         */
  	virtual int       getNumberOfHits(){return _hits.size();}

        /** Get the pointer to track hit list
         */
  	inline list<ModuleRowHit>& getHitsList()              {return _hits;}

        /** Plot hits of a track
         */
  	virtual void      drawTrack   (int color=2);

        /** Plot hits of a event
         */
	virtual void      drawEvent   (LCEvent* evt, string input, int color=1);

        /** Print track parameters
         */
	virtual void      printStateVector       ()=0;

        /** Set minimum number of hits on track 
         */
	void setMinTrackHits(int min){_minTrackHits= min;}

        /** Set maximum track skipping row
         */
 	void setMaxSkipRows (int max){_maxSkipRows = max;}

protected:
        /** Hit list of the track
         */
	list<ModuleRowHit>      _hits;

        /** An area on the TPC readout plane, now it is (module, row)
         */
	TPCHitDomain*           _domain;

        /** The container of all arranged hits
         */
	TPCHitMap*              _tpcHitMap;

        /** TPC geometry
         */
	const TPCParameters*    _tpcParameters;

        /** The last hit that is added to track
         */
	ModuleRowHit            _lastBestHit;

        /** Minimum number of hits on track
         */
	int _minTrackHits;

        /** Maximum track skipping row
         */
	int _maxSkipRows;
};

}
#endif  //ATTKALTRACK_H
