#ifndef HIT_TRACK_FINDER_TOPO_PROCESSOR_H
#define HIT_TRACK_FINDER_TOPO_PROCESSOR_H 1

#include "marlin/Processor.h"
#include "lcio.h"
#include <string>
#include <vector>
#include <IMPL/TrackerPulseImpl.h>

#include <gear/PadRowLayout2D.h>

#include "TopoFinder.h"

namespace marlintpc{

/** The HitTrackFinderTopoProcessor searches for hits and groups contiguous
 *  areas to track candidates.
 *  Using PadRowLayout2D from GEAR (RectangularPadRowLayout or FixedPadSizeDiskLayout)
 *  the agorithm searches for contiguous areas.
 *  Contiguous pulses in one pad row are grouped as hit candidates,
 *  then the hit coordinates are calculated using the centre of gravity method.
 *  The hits in one contiguous area are filled into a track candidate.
 *  \attention This processor only performs the search. The tracks are not seeded,
 *  i. e. the track paramaters of the track candidates are not set. They are just containers
 *  to hold the hits belonging to one track.
 *  
 *  If MarlinTPC is compiled with the $MARLIN_DEBUG option there is a lot of debug output.
 *  Use the steering file to reduce it / switch it off.
 *
 *  @author Martin Killenberg, University of Bonn
 *
 *  <h4>Input - Prerequisites</h4>
 *  \li A Collection of TrackerPulses
 *  \li GEAR TPCParameters. A zAnode user parameter should be set to get correct values. The
 *      maxDriftLength will be used as a fallback if not, but this might not be accurate.
 *  \li tpcconddata::TPCConditions for the drift velicity (optionally provided as processor parameter)
 *  \todo Handling of dead and noisy channels.
 *  \todo Improve z cut. Currently it does not use topological information which can cause problems with curlers.
 *
 *  <h4>Output</h4>
 *  \li Collection of Hits
 *  \li Collection of track candidates
 *
 *  @param InputTrackerPulses Name of input TrackerPulses collection (default: TPCPulses)
 *  @param OutputTrackerHits Name of output TrackerHits collection (default: TPCHits)
 *  @param OutputTrackCandidates Name of output Tracks collection (default: TPCTrackCandidates)
 *  @param InputTPCConditions \c OPTIONAL: Name of the input collection containing the TPC 
           conditions data (default: tpcconddata::TPCConditions::getDefaultColName() )
 *  @param SetOutputHitsTransient If not 0 the output collection is set transient (default: 0)
 *  @param SetOutputTrackCandidatesTransient If not 0 the output collection is set transient (default: 0)
 *  @param MaxEmptyPads Maximum number of consecuitve empty pads in cluster (default: 1)
 *  @param MinHitSize Minimum size of hit / Minimum number of Pads (default: 1)
 *  @param MaxTimeSpread Maximum time between pulses in a cluster [in ns] (default: 200)
 *  @param VDrift \c OPTIONAL: Set drift velocity in case there is no conditions data [in mm/us] (default: 0)
 *  @param NegativeZDirection Set to 1 for the second half TPC,
 *         which look into the negative z direction (default: 0)
 *
 *  The drift velocity should normally come from the conditions data. This can be  
 *  overwritten using this optional parameter. Do not set it in case you
 *  want to use vDrift from the CDDB!
 */

class HitTrackFinderTopoProcessor : public marlin::Processor , public TopoFinder {
  
 public:
  
  virtual Processor*  newProcessor() { return new HitTrackFinderTopoProcessor ; }
  
  
  HitTrackFinderTopoProcessor() ;
  
  /** Called at the begin of the job before anything is read.
   * Use to initialize the processor, e.g. book histograms.
   */
  virtual void init() ;
  
  /** Called for every run.
   */
  virtual void processRunHeader( EVENT::LCRunHeader* run ) ;
  
  /** Called for every event - the working horse.
   */
  virtual void processEvent( EVENT::LCEvent * evt ) ; 
  
  
  virtual void check( EVENT::LCEvent * evt ) ; 
  
  
  /** Called after data processing for clean up.
   */
  virtual void end() ;
  
  
 protected:

 private:
    /** the name of the collection the input pulses will be stored under
     */
    std::string _inputTrackerPulsesCollectionName ;

    /** the name of the collection the output hits will be stored under
     */
    std::string _outputTrackerHitsCollectionName ;

    /** the name of the collection the output tracks will be stored under
     */
    std::string _outputTracksCollectionName ;

    /** the name of the collection the TPC conditions are stored with
     */
    std::string _inputTPCConditionsCollectionName;

    /** Give the status which is set to the transient flag of the output collection.
     *  It is an int instead of a boolean, because the processor parameter cannot be a boolean.
     *  Hits and tracks can be set separately.
     */
    int _outputHitsTransient;    //< if not 0 the hits collection is set transient (default 0)
    int _outputTrackCandidatesTransient;  //< if not 0 the tracks collection is set transient (default 0)

//int _nRun ;
//int _nEvt ;
} ;

}//namespace marlintpc

#endif//HIT_TRACK_FINDER_TOPO_H
