#ifndef PedestalSubtractorProcessor_h
#define PedestalSubtractorProcessor_h 1

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "marlin/Exceptions.h"

#ifdef USE_LCCD
//LCCD
#include "lccd/ConditionsMap.hh"
#endif

// MarlinTPC
#include "Pedestal.h"
#include "ADCElectronicsHandler.h"

// C++
#include <string>


/// the MarlinTPC Package: reconstruction of TPC data
namespace marlintpc{

    /** Pedestal subtracting processor of the MarlinTPC package.
     * 
     *  Thisprocessor substracts pedestals provided via LCCD (simple file, database file or database).
     *
     *  <h4>Input - Prerequisites</h4>
     *  Collection of TrackerData
     *
     *  <h4>Output</h4> 
     *  Collection of TrackerData.
     * 
     * @param InputCollectionName Name of input data collection
     * @param OutputCollectionName Name of output data collection
     * @param SetOutputTransient If not 0 the output collection is set transient
     * @param PedestalCollectionName Name of the collection containing the pedestals
     * @param MaximumADCValue Maximum ADC Value of the read-out electronic to detect overflow samples, if this parameter is set it overides this information from the conditions database (optional parameter)
     **/
     /** @param ElectronicsParametersObjectName Name of the conditions object in which the parameters of the read out electronics is stored. For example "GenericADCElectronicsParameters". **/
     /** @param ElectronicsParametersCollectionName Name of the conditions collection in which the parameters of the read out electronics is stored **/
     /**
     *
     * @author M. E. Janssen, DESY; R. Diener, DESY
     */
                
     class PedestalSubtractorProcessor : public marlin::Processor {
    
        public:
        
        virtual Processor*  newProcessor() { return new PedestalSubtractorProcessor ; }
        
        PedestalSubtractorProcessor() ;
        
        virtual void init() ;
        
        virtual void processRunHeader( lcio::LCRunHeader* run ) ;
        
        virtual void processEvent( lcio::LCEvent * evt ) ; 
        
        virtual void check( lcio::LCEvent * evt ) ; 
        
        virtual void end() ;
        
        
        protected:
        
        /// the name of the input collection            
        std::string _inputColName ;    
        /// the name of the output collection
        std::string _outputColName ;  
        /// the name of the pedestal collection 
        std::string _pedestalColName ;
        /// the name of the collection with the read out electronics parameters
        std::string _electrParaObjName ;
        /// the name of the collection with the read out electronics parameters
        std::string _electrParaColName ;
    
        /** give the status which is set to the transient flag of the output collection
        *  it is an int instead of a boolean, because the processor parameter cannot be a boolean
        */
        int _outputIsTransient;
        
        /** is true till the first output collection is written to an event
        *  used to to store parameters only in the first collection
        *  (also done, if collection is transient)
        */
        bool isFirstCollection;
    
		/** maximum adc value, needed to determine if a pulse is maxed out
		 */
		int _maximum_ADC_value_override;

		/** The pedestal value to be used for every hardware channel in every readout group. This overrides any information provided by LCCD
		 */
		float _pedestal_value_override;
        
        private:

		#ifdef USE_LCCD
		/// the handler for run-time pedestal information
		typedef lccd::ConditionsMap< tpcconddata::Pedestal::key_type, tpcconddata::Pedestal> PedestalMap;
		PedestalMap *_pedestal_map;
		/// the handler for run-time electronics information
		typedef lccd::ConditionsMap< tpcconddata::GenericADCElectronicsParameters::key_type, tpcconddata::GenericADCElectronicsParameters> ElectronicParametersMap;
		ElectronicParametersMap *_electronic_parameters_map;
		#endif

     } ;
}

#endif



