#ifndef TRACK_FITTER_SIMPLE_CHI_SQUARE_H
#define TRACK_FITTER_SIMPLE_CHI_SQUARE_H

#include "TrackFitterBase.h"

namespace marlintpc
{

    /** Simple \f$\chi^2\f$ implementation of a track fitter.
     *
     *  It uses the getResiduals functions derrived from TrackFitterBase.
     *  The error  of each hit are calculated according to\n
     *  \f$\displaystyle \sigma_{\mathrm trans} = \sqrt{\sigma_{0,{\mathrm trans}}^2 + (D_{\mathrm trans} \sqrt{z})^2}\f$ \n
     *  \f$\displaystyle \sigma_{\mathrm long} = \sqrt{\sigma_{0,{\mathrm long}}^2 + (D_{\mathrm long} \sqrt{z})^2}\f$
     *
     *  The function minimised is
     *  \f$\displaystyle \chi^2 = \frac{res_{xy}^2}{\sigma_{\mathrm trans}} + \frac{res_{z}^2}{\sigma_{\mathrm long}}\f$
     */

class TrackFitterSimpleChiSquare : public TrackFitterBase
{
    public:
    /// Constructor for \f$\chi^2\f$ fitter, initializes the track fitter with the initial values
	TrackFitterSimpleChiSquare(double transDefocussing, double longDefocussing,
			     double transDiffusion, double longDiffusion);

	/// Well defined interface to create a new instance of the track fitter
	static TrackFitterBase * getInstance(LCParameters const * parameters);

	/// returns the revision of this implementation and of the TrackFitterBase it is derrived from
	virtual std::string getRevision();

	/// The actual fitter. Not implemented yet, thows gear::NotImplementedException
	virtual IMPL::TrackImpl *fitTrack(EVENT::Track const * seedTrack);

	/// returns "SimpleChiSquare"
	virtual unsigned char getFitterType() const;

	/// The transverse defocussing in the readout structure \f$\sigma_{0,{\mathrm trans}}\f$ (in mm)
	virtual double getTransDefocussing();

	/// The longitudinal defocussing in the readout structure \f$\sigma_{0,{\mathrm long}}\f$ (in mm)
	virtual double getLongDefocussing();

	/// The transverse diffusion coefficient (in mm/\f$\sqrt{\mathrm mm}\f$)
	virtual double getTransDiffusionCoef();

	/// The longitudinal diffusion coefficient (in mm/\f$\sqrt{\mathrm mm}\f$)
	virtual double getLongDiffusionCoef();

 protected:
	/// The defocussing and diffusion coefficients
	double  _transDefocussing, _longDefocussing, _transDiffusion, _longDiffusion;

 private:
	/// The private variable pointintg to the instance you get using getInstance function.
	/// This class is not explicitly a singleton, but if you use getInstance it behaves like one
	/// to avoid overhead by allocating and deallocating memory.
	static TrackFitterSimpleChiSquare * _theSimpleChiSquareFitter;
};

} // namespace marlintpc

#endif // TRACK_FITTER_CHI_SQUARE_H
