#ifndef TRACKMAKINGKALMANFILTERPROCESSOR_H
#define TRACKMAKINGKALMANFILTERPROCESSOR_H

#include <string>
#include <vector>
#include <map>

#include <EVENT/TrackerHit.h>
#include "marlin/Processor.h"
#include "lcio.h"
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/TrackerHitImpl.h>
#include <gear/PadRowLayout2D.h>
#include <gear/TPCParameters.h>

//Aida                        
#include <AIDA/AIDA.h>        
#include <marlin/AIDAProcessor.h>
#include <AIDA/ITupleFactory.h>
#include <AIDA/ITuple.h>

#include "AttKalTrack.h"


using namespace gear;
using namespace std;
using namespace EVENT;

namespace marlintpc
{
    /** This is processor which uses AttKalTrack class and
     *  its derived classes(i.e. StraightKalTrack and HeliKalTrack) for finding and fitting tracks.
     *  In those track classes,  Kalman filter algorithm and track hits are included.
     *  
     *  @author Bo Li, Tsinghua University
     *  @author Keisuke Fujii, KEK
     *
     *  <h4>Input - Prerequisites</h4>
     *  Collection of TrackerHits
     *
     *  <h4>Output</h4> 
     *  Collection of Tracks
     * 
     *  @param InputTrackerHits Name of the Input TrackerHits collection
     *  @param OutputTracks Name of the output Tracks collection
     *  @param MinTrackHits Minimum number of hits on track (default: 15)
     *  @param MaxSkipRows Maximum number of subsequnetly missing hits (default: 5)
     *  @param TrackType Track type (default: 1, 0=straight line 1=helix)
     */

class TrackMakingKalmanFilterProcessor : public marlin::Processor {
  
public:
	TrackMakingKalmanFilterProcessor();

	~TrackMakingKalmanFilterProcessor();

	Processor * newProcessor();

        /** Called at the begin of the job before anything is read.
         */
	void init();

        /** Called for every run.
         */
	void processRunHeader (LCRunHeader *run);
	
        /** Called for every event - the working horse.
         */
	virtual void processEvent(LCEvent *evt);

	virtual void check (LCEvent *evt);

        /** Called after data processing for clean up.
         */
	virtual void end();

private:
        /** Name of the input TrackerHitsCollection
         */
	std::string _InputTrackerHitsCollectionName;

        /** Name of the outout TracksCollection
         */
	std::string _outputTracksCollectionName;
        
        /** The minimum hit number to discriminate a track 
         */
	int _minTrackHits;

        /** The maximum step over which a track can skip for one time
         */
	int _maxSkipRows;

        /** The total row number for a module
         */
	int _nRow;

        /** Track type: Straight line(0) or helix(1)
         */
	int _trackType;

        /** The order for picking up hits from modules
         */
	map<int,int>*               _moduleOrder;

        /** The pointer to a TPCHitMap which arranges hits by module and row
         */
	TPCHitMap*                  _hitMap;

        /** The geometry pointer for TPC
         */
	const TPCParameters*        _tpcParameters;

        /** The track collection in a event
         */
	LCCollectionVec*            _tpcTracksCollection;

	/** Tuple for track parameters
	 */
	AIDA::ITuple*               _trackTuple;

        /** The point to the track class which is used for tracking
         */
	static AttKalTrack*         _theKalTrackPtr;
};

}// namespace marlintoc

#endif // TRACKMAKINGKALMANFILTERPROCESSOR_H
