#include "PedestalSubtractorProcessor.h"

// Marlin
#include <marlin/ConditionsProcessor.h>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/LCFlagImpl.h>
#include <UTIL/LCFixedObject.h>

// tpcconditions
#include "GenericADCElectronicsParameters.h"

// C++
#include <iostream>

using namespace lcio ;
using namespace marlin ;
using namespace tpcconddata;
using namespace std;

namespace marlintpc
{

    PedestalSubtractorProcessor aPedestalSubtractorProcessor ;

    PedestalSubtractorProcessor::PedestalSubtractorProcessor() : Processor("PedestalSubtractorProcessor")
    {

        /* description of the processor which will be displayed in the
         * steering file autometicly generated by
         * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
         */
        _description = "PedestalSubtractorProcessor to correct the TrackerData by subtracting the pedestals provided via LCCD";

        /* register steering parameters: name, description, class-variable, default value
         * the type will by definded by the type of the default value
         * string, double, float and int are possible
         * plaese note to get a float you must cast in C++: (float) 1.3
         */
        registerInputCollection( LCIO::TRACKERDATA,
                                 "InputCollectionName" ,
                                 "Name of the input collection"  ,
                                 _inputColName ,
                                 std::string("TPCConvertedRawData") 
                               );

        registerOutputCollection( LCIO::TRACKERDATA,
                                  "OutputCollectionName" ,
                                  "Name of the output collection"  ,
                                  _outputColName ,
                                  std::string("TPCData") 
                                );

        registerProcessorParameter( "SetOutputTransient" ,
                                    "if not 0 the output collection is set transient"  ,
                                    _outputIsTransient,
                                    1
                                  );

        registerProcessorParameter( "PedestalCollectionName" ,
                                    "Name of the collection containing the pedestals",
                                    _pedestalColName,
                                    tpcconddata::Pedestal::getDefaultColName()
                                  );
        
        registerProcessorParameter( "ElectronicsParametersObjectName" ,
                                    "Name of the conditions object in which the parameters of the read out electronics is stored",
                                    _electrParaObjName,
                                    std::string("GenericADCElectronicsParameters")
                                  );
        
        registerProcessorParameter( "ElectronicsParametersCollectionName" ,
                                    "Name of the conditions collection in which the parameters of the read out electronics is stored",
                                    _electrParaColName,
                                    tpcconddata::GenericADCElectronicsParameters::getDefaultColName() 
                                  );

		registerOptionalParameter("MaximumADCValueOverride",
				"Maximum number of the adc spectrum; to determine wether a pulse is maxed out. This single value overrides any value, from any readout group, which may have been taken from LCCD.",
				_maximum_ADC_value_override,
				255 );

		registerOptionalParameter("PedestalValueOverride",
				"The pedestal value to be used for every hardware channel in every readout group. This overrides any information provided by LCCD.",
				_pedestal_value_override,
				(float)0. );


    }

    void PedestalSubtractorProcessor::init()
    {
        printParameters() ;

		// check if the LCCD pedestal handler will be used
		if ( parameterSet("PedestalValueOverride") ) {

			m_out(WARNING)<<"Pedestal value ("<<_pedestal_value_override<<") taken from the Marlin steering file. This overrides any LCCD information."<<std::endl;
			_pedestal_map = NULL;
		}
		else {

			// the pedestal map which provides updated run-time pedestal information
			_pedestal_map = new PedestalMap( &Pedestal::conditions_key );
			ConditionsProcessor::registerChangeListener( _pedestal_map , _pedestalColName ) ;
		}
        
        // give warning message if optional maxADCValue parameter overrides conditions data
        if ( parameterSet("MaximumADCValueOverride") ) 
        {
            m_out(WARNING)<<"Maximum ADC value ("<<_maximum_ADC_value_override<<") taken from steering file. This overrides parameter from conditions database!"<<std::endl;
			_electronic_parameters_map = NULL;
        }
        else
        {
			_electronic_parameters_map = new ElectronicParametersMap( &GenericADCElectronicsParameters::conditions_key );
			ConditionsProcessor::registerChangeListener( _electronic_parameters_map , _electrParaColName ) ;
        }

    }
    
    
    void PedestalSubtractorProcessor::processRunHeader( LCRunHeader* run)
    {
        run->parameters().setValue( _processorName + "_revision", "$Rev: 2197 $");
        for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ )
        {
            if ( ! i->second->isOptional() || i->second->valueSet() )
            {
                run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
            }
        }
    }

    void PedestalSubtractorProcessor::processEvent( LCEvent * evt )
    {
        // get the input collection from the event
        try
        {
            LCCollection* inputCol = evt->getCollection( _inputColName ) ;

			m_out(MESSAGE) << "Processing " << inputCol->getNumberOfElements() << " TrackerData elements from collection " << _inputColName << "." << std::endl;
            // create the output collection with a string defining the TYPE (see lcio.h)
            LCCollectionVec* outputCol = new LCCollectionVec( LCIO::TRACKERDATA );
            // set the transient flag as definded by the processor parameter
            outputCol->setTransient( (_outputIsTransient != 0) );
            // set the flags that cellID1 should be stored
            lcio::LCFlagImpl trkFlag(0) ;
            trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
            outputCol->setFlag( trkFlag.getFlag() );
            
            // get the number of elements in the input collection
            int nRawData = inputCol->getNumberOfElements();
            // loop over the number of elements of the input collection
            for(int iRawData=0; iRawData< nRawData ; iRawData++)
            {
                /* getElement return only a LCObject
                 * you have to cast it the get the right type
                 * please use dynamic_cast which will return 0 if the object can not be casted
                 */
                TrackerData* RawData = dynamic_cast<TrackerData*>( inputCol->getElementAt( iRawData ) ) ;

				//only poceed if the cast to TrackerData was successful
				if ( RawData == 0 ) {

					continue;
				}

				// create a new object of the TrackerData type
				TrackerDataImpl* Data = new TrackerDataImpl();

				// transver the values from the "RawData" to the "Data" object

				// first the simple
				Data->setCellID0( RawData->getCellID0() );
				Data->setCellID1( RawData->getCellID1() );
				Data->setTime( RawData->getTime() );

				m_out(DEBUG)<<"Processing Spectrum at Pad:"<<RawData->getCellID0()<<"/Module:"<<RawData->getCellID1()<<std::endl;

				// get the pedestal
				float PedestalVal = 0.;
				if ( parameterSet("PedestalValueOverride") ) {

					PedestalVal = _pedestal_value_override;
				}
				else {

					#ifdef USE_LCCD
					// the key is currently (hardware channel, hardware group)
					PedestalVal = _pedestal_map->find( Pedestal::key_type( Data->getCellID0(), Data->getCellID1() ) ).getPedestalValue();
					#endif
				}

				m_out(DEBUG)<<"Pedestal Value here is: "<<PedestalVal<<std::endl;
				FloatVec RawDataADCVal = RawData->getChargeValues();
				m_out(DEBUG)<<"Values (before/after) : ";

				// read in the maximum possible ADC value, either from LCCD conditions if parameter not set
				int maximum_adc_value = 0;
				if ( parameterSet("MaximumADCValueOverride") ) {

					maximum_adc_value = _maximum_ADC_value_override;
				}
				else {

					maximum_adc_value = _electronic_parameters_map->find( Data->getCellID1() ).getMaxADCValue();
				}

				for ( FloatVec::iterator iSlice=RawDataADCVal.begin();
						iSlice!=RawDataADCVal.end(); iSlice++ )
				{
					float outSlice=0.0;
					if ( (*iSlice)==maximum_adc_value ) // test if slice value is maximum of read-out electronics == overflow
					{
						outSlice = (*iSlice); // if slice in overflow, do not substract pedestal to preserve the overflow information
						m_out(DEBUG)<<" ! Slice not pedestal substracted: ";
					}
					else if ( (*iSlice) > maximum_adc_value )
					{
						m_out(ERROR)<<std::endl<<"Error: Value of time slice (="<<(*iSlice)<<") is bigger than maximum ADC value (="<<maximum_adc_value<<") !"<<std::endl;
						throw StopProcessingException(this);
					} else {
						outSlice=(*iSlice)-PedestalVal; // else substrect pedestal value
					}
					// debug output
					m_out(DEBUG)<<(*iSlice)<<"/"<<outSlice<<" ";
					// fill the data vector with the pedestal substracted data
					(Data->chargeValues()).push_back( outSlice );
				}
				m_out(DEBUG)<<std::endl<<std::endl;
				/* fill the TrackerData in the collection
					* please note: DO NOT DELETE the object! this will be done by LCIO
					*/
				outputCol->addElement( Data );

            } // end of loop over the number of elements of the input collection

            // save the processor parameters in the first collection stored
            if (isFirstCollection)
            {
                outputCol->parameters().setValue("InputCollectionName",_inputColName);
                // TO DO: make this sense?
                outputCol->parameters().setValue("SetOutputTransient",_outputIsTransient);
                outputCol->parameters().setValue("PedestalCollectionName",_pedestalColName);
                #ifndef USE_LCCD
                outputCol->parameters().setValue("PedestalFileName",_pedestalFileName);
                #endif
                isFirstCollection = false;
            }

            /* after converting all raw data objects
             * add the collection to the event and asigning a name to it
             * please note again: DO NOT DELETE the collection  this will be done by LCIO
             */
            evt->addCollection( outputCol, _outputColName );

        }
        catch(DataNotAvailableException)
        {
            // if there is not input collection give a warning to the user
            std::cerr<<"Warning of PedestalSubtractorProcessor::processEvent: input collection "
                     <<_inputColName<<" cannot be found in this event." << endl;
        }
    }

    void PedestalSubtractorProcessor::check( LCEvent * evt )
    {
        // nothing here currently - could be extended to fill checkplots
    }

    void PedestalSubtractorProcessor::end() {

		if ( _pedestal_map ) {

			delete _pedestal_map;
			_pedestal_map = NULL;
		}

		if ( _electronic_parameters_map ) {

			delete _electronic_parameters_map;
			_electronic_parameters_map = NULL;
		}
		
	}

}
