#include "PulseFinderProcessor.h"

// tpcconditions
#include "GenericADCElectronicsParameters.h"

// Marlin
#include <marlin/ConditionsProcessor.h>

// gsl
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_multifit.h>

//LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/LCFlagImpl.h>

#ifdef USE_GEAR
// STUFF needed for GEAR
#include <marlin/Global.h>
#include <gear/GEAR.h>
#include <gear/TPCParameters.h>
#endif

//C++
#include <iostream>
#include <algorithm>
#include <iterator>
#include <cmath>

using namespace lcio;
using namespace marlin;
using namespace std;
using namespace tpcconddata;

namespace marlintpc
{
  PulseFinderProcessor aPulseFinderProcessor;
  
  PulseFinderProcessor::PulseFinderProcessor() : Processor("PulseFinderProcessor") {

    _description = "PulseFinderProcessor: finds pulses out of the raw data and calculate the time and charge (in FADC) information";
    
    registerInputCollection(LCIO::TRACKERDATA,
			    "InputCollectionName",
			    "Name of the input TrackerData collection",
			    _inputColName,
			    string("TPCData"));
    
    registerOutputCollection(LCIO::TRACKERDATA,
			     "OutputTrackerDataCollectionName",
			     string("Name of the output TrackerData collection ")+
			     string("only the data related to a pulse will be stored"),
			     _outputDataColName,
			     string("TPCDataInPulses"));
	
    registerOutputCollection(LCIO::TRACKERPULSE,
			     "OutputTrackerPulseCollectionName",
			     "Name of the output TrackerPulse collection",
			     _outputPulseColName,
			     string("TPCPulses"));

    registerProcessorParameter("SetOutputTransient",
			       "if true the output collection is set transient",
			       _outputIsTransient,
			       true);
	
    registerProcessorParameter("ForceSpectrumSave",
			       "if true the ADC spectrum will be stored for all pulses",
			       _forceSpectrumSave,
			       true);
	
    registerProcessorParameter("PulseStartThreshold",
			       string("A pulse starts if the ADC value is the given value times the pedestal width."),
			       _pulseStartThreshold,
			       (float)7.0);

    registerProcessorParameter("PulseEndThreshold",
			       string("A pulse ends if the ADC value drops below this value times the pedestal width."),
			       _pulseEndThreshold,
			       (float)3.0);

    registerProcessorParameter("MinimumPulseHeight",
			       string("Declares the minimal height of the maximum entry of a pulse in ADC counts."),
			       _minimumPulseHeight,
			       (float)12.0);
    
    registerProcessorParameter("MinimumPulseLength",
			       string("Declares the minimal length of a pulse in time bins above the thresholds."),
			       _minimumPulseLength,
			       5);

    registerProcessorParameter("SaveNBinsBeforeStart",
			       "# bins saved before start threshold (except there is another pulse)",
			       _saveNBinsBeforeStart,
			       2);
	
    registerProcessorParameter("SaveNBinsAfterEnd",
			       "# bins saved after the end threshold (except there is another pulse)",
			       _saveNBinsAfterEnd,
			       8);

	registerProcessorParameter( "ElectronicsParametersObjectName" ,
			"Name of the conditions object in which the parameters of the read out electronics is stored",
			_electrParaObjName,
			std::string("GenericADCElectronicsParameters")
			);

	registerProcessorParameter( "ElectronicsParametersCollectionName" ,
			"Name of the conditions collection in which the parameters of the read out electronics is stored",
			_electrParaColName,
			tpcconddata::GenericADCElectronicsParameters::getDefaultColName()
			);

	registerOptionalParameter( "UseParabolicFitAmplitudeAsCharge?" ,
			"Set this flag to use the amplitude calculated by fitting a parabola to the log of the pulse data.",
			_use_fit_charge,
			false
			);

    registerOptionalParameter("UseParabolicFitMaxAsTime?",
			      "Set this flag to use the time caclulated by fitting a parabola to the log of the pulse data.",
			      _use_fit_time,
			      false);

	// now some parameters for overriding values which should be obtained from LCCD
	registerOptionalParameter("MaximumADCValueOverride",
				"Maximum number of the adc spectrum; to determine wether a pulse is maxed out. This single value overrides any value, from any readout group, which may have been taken from LCCD.",
				_maximumADCValueOverride,
				(float)255.);

	registerOptionalParameter("PedestalWidthOverride",
				"The pedestal width to be used for every hardware channel in every readout group. This overrides any information provided by LCCD.",
				_pedestal_width_override,
				(float)0.);

	registerOptionalParameter("SignalPolarityOverride",
				"The polarity of the signals (1 for positive, -1 for negative) produced by all readout groups. This overrides any information provided by LCCD.",
				_polarity_override,
				1 );

    registerOptionalParameter("ReadoutFrequencyOverride",
				"The readout frequency of every readout group (in Hz). This overrides any information provided by LCCD.",
				_readout_frequency_override,
				(float)1.E9);
	// end of the override variables

    registerProcessorParameter("PedestalCollectionName",
			       "Name of the collection containing the pedestal information",
			       _pedestalColName,
			       Pedestal::getDefaultColName() );
	
    /* if MARLIN is compiled without LCCD a file is needed which contains the pedestal information
     */
#ifndef USE_LCCD
    registerProcessorParameter("PedestalCollectionFile",
			       "File containing the pedestals",
			       _pedestalFileName,
			       string("Pedestal.slcio"));

	_pedestalHandler = NULL;
	_adcElectronicsHandler = NULL;
#else
	_pedestal_map = NULL;
	_electronic_parameters_map = NULL;
#endif
  }



void PulseFinderProcessor::init()
  {
    
    // usually a good idea to
    printParameters();
    
    _nRun = 0;
    _nEvt = 0;

    // if Marlin is compiled with GEAR the readout frequency will be given via the GEAR steering file
#ifdef USE_GEAR
    _readout_frequency_override = Global::GEAR->getTPCParameters().getReadoutFrequency();
#endif
    
    // determine if the pedestal information is supplied by LCCD
	if( !parameterSet("PedestalWidthOverride") ) {

		#ifdef USE_LCCD
		_pedestal_map = new PedestalMap( &Pedestal::conditions_key );
		ConditionsProcessor::registerChangeListener( _pedestal_map , _pedestalColName ) ;
		#else
		streamlog_out( DEBUG ) << "HERE" << endl;
		//_pedestalHandler = new PedestalHandler( _pedestalColName, _pedestalFileName );
		// create the object for handling the electronics parameters
		_adcElectronicsHandler = new ADCElectronicsHandler( _electrParaObjName, _electrParaColName );
		#endif
	}

	// determine if the polarity and readout frequency are supplied by LCCD
	if( !parameterSet("ReadoutFrequencyOverride") && !parameterSet("SignalPolarityOverride") ) {

		_electronic_parameters_map = new ElectronicParametersMap( &GenericADCElectronicsParameters::conditions_key );
		ConditionsProcessor::registerChangeListener( _electronic_parameters_map , _electrParaColName ) ;
	}
}



void PulseFinderProcessor::processRunHeader(LCRunHeader* run)
  {
    run->parameters().setValue( _processorName + "_revision", "$Rev: 2210 $");
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ )
      {
	if ( ! i->second->isOptional() || i->second->valueSet() )
	  {
	    run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
	  }
      }
    
    _nRun++;
  }



void PulseFinderProcessor::processEvent(LCEvent* evt)
  {
    // give a life sign if you are in debug mode
    _nEvt++;
    m_out(MESSAGE) << "Finding the pulses in " <<_inputColName<< "." << endl;
    
    try
      {
	LCCollection* InputCollection = evt->getCollection(_inputColName);
	
	// create the output collections for pulses and zero suppressed data (ADC spectra)
	IMPL::LCCollectionVec* outputColPulse = new IMPL::LCCollectionVec(LCIO::TRACKERPULSE);
	IMPL::LCCollectionVec* outputColData  = new IMPL::LCCollectionVec(LCIO::TRACKERDATA);
	
	// set the transient flag as definded by the processor parameter
	outputColPulse->setTransient(_outputIsTransient);
	outputColData->setTransient(_outputIsTransient);
	
	// set the flags that cellID1 should be stored
	lcio::LCFlagImpl trkFlag(0);
	trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1);
	outputColPulse->setFlag(trkFlag.getFlag());
	outputColData->setFlag(trkFlag.getFlag());
	
	/* temporary data object, to be filled by /raw/ pulse finding 
  	   first pair object is the actual tracker data,
	   second pair object is the quality
	*/
	vector<pair<TrackerDataImpl*, int> > tempPulses;
	tempPulses.clear();

	unsigned int nData = InputCollection->getNumberOfElements();

	// first loop over all channels within the event to search for the /raw/ pulses
	for(unsigned int iData = 0; iData < nData; ++iData)
	  {
	    TrackerData* thisData = dynamic_cast<TrackerData*>( InputCollection->getElementAt(iData) );

	    if (thisData)    // cast is OK
	      {

			#ifdef USE_LCCD
			m_out(DEBUG) << "Looking at Pad:" << thisData->getCellID0()
			     << " module:" << thisData->getCellID1() << endl;
			m_out(DEBUG) << "Pedestal Width for pad:" << thisData->getCellID0()
			     << " module:" << thisData->getCellID1() << " = "
			     << get_pedestal_width( thisData->getCellID0(), thisData->getCellID1() ) << endl;

			float noiseValue = get_pedestal_width( thisData->getCellID0(), thisData->getCellID1() ) * _pulseStartThreshold;
			#else
			m_out(DEBUG) << "Looking at Pad:" << thisData->getCellID0()
			     << " module:" << thisData->getCellID1() << endl;
			m_out(DEBUG) << "Pedestal Value for pad:" << thisData->getCellID0()
			     << " module:" << thisData->getCellID1() << " = "
			     << _pedestalHandler->getPedestalValue( thisData->getCellID0(), thisData->getCellID1() ) << endl;
			m_out(DEBUG) << "Pedestal Width for pad:" << thisData->getCellID0()
			     << " module:" << thisData->getCellID1() << " = "
			     << _pedestalHandler->getPedestalWidth( thisData->getCellID0(), thisData->getCellID1() ) << endl;

			float noiseValue = _pedestalHandler->getPedestalWidth(thisData->getCellID0(), thisData->getCellID1()) * _pulseStartThreshold;
			#endif

		// if pedestal width is undefined, its default value is -1
		//  then use pedestal width = 1
		if (noiseValue < 0.)
		  noiseValue = _pulseStartThreshold;
		
		// this is the master call: search in the spectrum for pulses and
		// store them into the collection vector that is passed to the call
		findPulses(thisData, &tempPulses, noiseValue);
	      }
	  }

	// use the found pulses to calculate time and charge information and store the data
	for(vector<pair<TrackerDataImpl*, int> >::const_iterator pulseCandidate = tempPulses.begin();
	    pulseCandidate < tempPulses.end(); ++pulseCandidate)
	  {
	    TrackerDataImpl* trackerData = pulseCandidate->first;
	    int quality                  = pulseCandidate->second;

	    //save Pulse
	    TrackerPulseImpl* newPulse = new TrackerPulseImpl();
	    newPulse->setCellID0(trackerData->getCellID0());
	    newPulse->setCellID1(trackerData->getCellID1());

		pair<float, float> charge_and_time = calculate_charge_and_time( trackerData );
	    newPulse->setCharge( charge_and_time.first );
		newPulse->setTime( charge_and_time.second );

	    newPulse->setQuality(quality);

		if( streamlog::out.write< streamlog::DEBUG >() ) {

			m_out(DEBUG) << "[***] Found a pulse on [Pad/Module]: [" << trackerData->getCellID0()
					<< "/" << trackerData->getCellID1() << "], at the time " << newPulse->getTime() << " of length "
					<< (trackerData->getChargeValues()).size() << " with quality flag \"" << quality << "\" containing charge " << newPulse->getCharge() << '.' << endl;

			m_out(DEBUG) << "[***] The pulse spectrum is : ";
			
			vector<float> theADCSpectrum = trackerData->getChargeValues();
			for(vector<float>::const_iterator adcVal = theADCSpectrum.begin(); adcVal < theADCSpectrum.end(); ++adcVal)
			{
				m_out(DEBUG) << *adcVal << " ";
			}
			m_out(DEBUG) << endl;
		}
	    
	    // store the adc spectrum connected to the pulse
	    // this is implicitly done by incorporating /_forceSpectrumSave/ into /quality/
	    if ( quality != 0 || _forceSpectrumSave) {

			// make a copy of the TrackerData
			TrackerDataImpl* newData = new TrackerDataImpl();
			newData->setCellID0(trackerData->getCellID0());
			newData->setCellID1(trackerData->getCellID1());
			newData->setTime(trackerData->getTime());
			vector<float> thisDataChargeVal = trackerData->getChargeValues();
			for (vector<float>::const_iterator iSlice = thisDataChargeVal.begin(); iSlice != thisDataChargeVal.end(); ++iSlice)
			(newData->chargeValues()).push_back( (*iSlice) );
		
			// add the spectrum to the <tracker data> collection
			outputColData->addElement(newData);
			// and add the tracker data to the <tracker pulse> collection
			newPulse->setTrackerData(newData);
	      }

	    outputColPulse->addElement(newPulse);
	  }
	
	evt->addCollection(outputColPulse, _outputPulseColName);
	evt->addCollection(outputColData, _outputDataColName);

	m_out(MESSAGE) << "Found " << outputColPulse->getNumberOfElements() << " pulses." << endl;
      }
    catch(DataNotAvailableException)
      {
	// if there is not input collection give a warning to the user
	cerr << "Warning of PulseFinderProcessor::processEvent: input collection "
	     << _inputColName << " cannot be found in event " << _nEvt << endl;
      }
  }



void PulseFinderProcessor::check(LCEvent* evt)
  {
  }



void PulseFinderProcessor::end() {

	#ifdef USE_LCCD
	if( _pedestal_map ) {

		delete _pedestal_map;
		_pedestal_map = NULL;
	}
	if( _electronic_parameters_map ) {

		delete _electronic_parameters_map;
		_electronic_parameters_map = NULL;
	}
	#else
	if( _pedestalHandler ) {

		delete _pedestalHandler;
		_pedestalHandler = NULL;
	}
	#endif
}



void PulseFinderProcessor::findPulses(TrackerData* thisData, vector<pair<TrackerDataImpl*,int> >* tempPulses, float noiseValue)
  {
    /** the basic idea to find pulses is split into three steps:
       (1) simply look for candidate pulses: above values between start and end threshold 
          due to the high frequency of this already check here for a minimum length
       (2) search the candidate pulses for possible multiple pulse and split, if necessary
       (3) finally check the pulse boundaries and assemble the pulse out of the bins before and after the pulse, if possible
    */

	// get the maximum ADC value
	const float maximumADCValue = get_max_ADC_value( thisData->getCellID1() );

	// determine the polarity of the data
	const int polarity = get_polarity( thisData->getCellID1() );

    float startTime  = thisData->getTime(); 

    vector<float> ADCSpectrum = thisData->getChargeValues();
    
    /* this is a help construct for the data (including a typedef)
       simply store the start and end position of the pulse
    */
    typedef std::vector<float>::iterator timeIter;
    vector<pair<timeIter, timeIter> > theCandidatePulses;
    
    /* do the first task: 
       simply look for the values above the start threshold and below end threshold
    */
    brutePulseSearch(ADCSpectrum, &theCandidatePulses, polarity);

    /* do the check for multiple pulses,
       technically: divide the start- and end-iterator into a new pair
       TODO: in addition to pulse splitting adjust the charge at the boundaries
    */
    //splitMultiplePulses(&theCandidatePulses, noiseValue);
    
    /* as last part determine the possible boundaries of the pulse;
       - check wether the storing of the time bins before and after the end is possible
       - set the flags accordingly
    */
    
    /* now prepare the pulse candidates for storage:
       - assign quality flags (FIXME: definition still to be done!)
       - check boundaries at the beginning and the end
       - return the data construct for storage
    */

	for(vector<pair<timeIter, timeIter> >::const_iterator pulse_boundaries = theCandidatePulses.begin(); pulse_boundaries < theCandidatePulses.end(); pulse_boundaries++) {

		int quality = 0;

		// check if we're on the first or last pulse boundaries
		const timeIter previous_pulse_end = (pulse_boundaries == theCandidatePulses.begin())? ADCSpectrum.begin(): (pulse_boundaries - 1)->second;
		const timeIter next_pulse_start = (pulse_boundaries == (theCandidatePulses.end()-1))? ADCSpectrum.end(): (pulse_boundaries + 1)->first;

		// check for a double-pulse candidate (overlapping boundaries)
		if( (distance( previous_pulse_end, pulse_boundaries->first ) < 0) || (distance( pulse_boundaries->second, next_pulse_start ) < 0) ) {

			// bit 0: pulse is a double-pulse candidate
			quality |= 0x01;
		}

		// check for anomalous pulse shape
		//FIXME: perhaps an anomolous pulse is best explained in terms of processor paramters. for example
		//	bool isShortTailAnomalous? = bool
		//  bool isLongTailAnomalous? = bool
		//   int anomalousTailLength = nn
		// float anomalousLengthHeighRatio = nn.nn
		if( false ) {

			// bit 2: anomalous pulse shape
			quality |= 0x04;
		}

		// check for underflow
		//FIXME: what is underflow? surely there will be some bin values of "zero" charge in simulations. perhaps this only makes sense when noise is included in the channel
		if( false ) {

			// bit 3: underflow in at least one time bin
			quality |= 0x08;
		}

		// check for overflow
		const timeIter max = max_element( pulse_boundaries->first, pulse_boundaries->second );

		if( *max >= maximumADCValue ) {

			// bit 4: overflow in at least one time bin
			quality |= 0x10;
		}

		// now copy the data to a TrackerData object
		timeIter copy_start = pulse_boundaries->first;
		const int start_displacement = std::max( -_saveNBinsBeforeStart, (int)distance( copy_start, previous_pulse_end ) );
		advance( copy_start, start_displacement );

		timeIter copy_end = pulse_boundaries->second;
		advance( copy_end, min( _saveNBinsAfterEnd, (int)distance( copy_end, next_pulse_start ) ) );

		// calculate the start time of the pulse to be stored:
		float currentTime = startTime + (pulse_boundaries->first - ADCSpectrum.begin()) - start_displacement;

		TrackerDataImpl* toStoreTrackerData;
		toStoreTrackerData = new TrackerDataImpl();
		toStoreTrackerData->setCellID0(thisData->getCellID0());
		toStoreTrackerData->setCellID1(thisData->getCellID1());
		toStoreTrackerData->setTime(currentTime);
		toStoreTrackerData->chargeValues().resize( distance( copy_start, copy_end ) + 1 );

		copy( copy_start, copy_end, toStoreTrackerData->chargeValues().begin() );

		pair<TrackerDataImpl*, int> tempPair( toStoreTrackerData, quality );
		tempPulses->push_back(tempPair);
	}

    return;
}


void PulseFinderProcessor::brutePulseSearch(vector<float>& ADCSpectrum, vector<pair<vector<float>::iterator, vector<float>::iterator> >* theCandidatePulses, const int polarity) {

	typedef vector<float>::iterator timeIter;
	timeIter candStart, candEnd;

	// determine if the pulse is negative
	const bool negativePolarity = (polarity == -1);

    // also look if the pulse candidate meets the minimum requirements of length and height
	// loop through the pulse data looking for continuous areas of charge

	vector<float>::iterator current_value = ADCSpectrum.begin();

	int candidate_length = 0;
	float candidate_max_height = 0.;
	bool inside_pulse = false;

	while( current_value < ADCSpectrum.end() ) {

		if( !inside_pulse ) {

			// we're not inside a pulse, look for the start
			const bool above_start_threshold = negativePolarity? (*current_value <= _pulseStartThreshold): (*current_value >= _pulseStartThreshold);

			if( above_start_threshold ) {

				inside_pulse = true;
				candidate_length = 1;
				candidate_max_height = *current_value;
				candStart = current_value;
			}
		}
		else {

			// we're inside a pulse, look for the end
			const bool below_end_threshold = negativePolarity? (*current_value > _pulseEndThreshold): (*current_value < _pulseEndThreshold);
			const bool end_of_data = (current_value + 1 == ADCSpectrum.end());

			if( !below_end_threshold ) {

				++candidate_length;
				candidate_max_height = negativePolarity? min( *current_value, candidate_max_height ): max( *current_value, candidate_max_height );
			}

			if( below_end_threshold || end_of_data ) {

				// determine if this pulse is a candidate
				const bool meets_height_requirement = negativePolarity? (candidate_max_height <= _minimumPulseHeight): (candidate_max_height >= _minimumPulseHeight);

				const bool meets_length_requirement = (candidate_length >= _minimumPulseLength);

				if( meets_height_requirement && meets_length_requirement ) {

					m_out(DEBUG1) << "Candidate accepted. Reason?: max height " << candidate_max_height << ", length " << candidate_length << "." << endl;
					theCandidatePulses->push_back( pair<timeIter,timeIter>(candStart,current_value) );
				}
				else {

					// this feels like fishing, the pulse wasn't big enough so toss it back
					m_out(DEBUG1) << "Candidate rejected. Reason?: below height " << !meets_height_requirement << ", below length " << !meets_length_requirement << "." << endl;
				}

				// reset the candidate
				inside_pulse = false;
			}
		}

		++current_value;
	}
}


  void PulseFinderProcessor::splitMultiplePulses(vector<pair<vector<float>::iterator, vector<float>::iterator> >* theCandidatePulses, float noiseValue)
  {
    typedef vector<float>::iterator timeIter;

    /* now check the candidates for multiple pulses
       since the iterators would be invalidated do this in two steps:
       first find the positions where the candidate needs to be split
        for this: construct the derivative vector for the pulse in question
       then do the actual splitting
    */

    // create a temporary storage for the "new" separated pulses
    vector<pair<timeIter, timeIter> > tempPulseStorage;
    bool needsSplitting = false;
    
    for(vector<pair<timeIter, timeIter> >::iterator aPulse = theCandidatePulses->begin();aPulse < theCandidatePulses->end(); ++aPulse)
      {
	vector<float> derivativeVec;
	vector<unsigned int> splitPositions;
	
	/* for each candidate in the lot:
	   loop from start to end of pulse and check wether
	   a change in the derivative from positive to negative occurs 
	*/
	for( timeIter aBin = aPulse->first; aBin < (aPulse->second - 1); ++aBin)
	  {
	    double derivative = *(aBin + 1) - *aBin;
	    derivativeVec.push_back(derivative);
	  }
	
	// now take a closer look at the candidate
	for(vector<float>::const_iterator dVal = derivativeVec.begin(); dVal+1 < derivativeVec.end() ; ++dVal)
	  {
	    // check the sign change and if the absolute change is above the noise threshold
	    // store the position of the /end/ of the split pulse + 1 (this is the split value +2)
	    if( (*dVal < 0) && (*(dVal+1) > 0) && (fabs(*dVal - *(dVal+1)) > noiseValue) )
	      {
		splitPositions.push_back(dVal-derivativeVec.begin() + 2);
		needsSplitting = true;
		/* still under construction:
		   for a more exact charge evaluation the charge needs to be spread between the split pulses
		   
		   double factor_change = *(dVal+1) / (*dVal - *(dVal + 1) );
		*/
	      }
	  }

	if(splitPositions.empty())
	  tempPulseStorage.push_back(*aPulse);
	else
	  {
	    timeIter currentStart = aPulse->first;
	    pair<timeIter, timeIter> tmpPair;

	    // do the splittings
	    for(vector<unsigned int>::const_iterator split = splitPositions.begin(); split < splitPositions.end(); ++split)
	      {
		tmpPair.first  = currentStart;
		tmpPair.second = aPulse->first + *split;
		currentStart   = aPulse->first + *split;
		tempPulseStorage.push_back(tmpPair);
	      }
	    // and store the trailing end
	    tmpPair.first  = currentStart;
	    tmpPair.second = aPulse->second;
	    tempPulseStorage.push_back(tmpPair);
	  }
      }
    // replace the original vector with the new one
    if(needsSplitting)
      {
	for(vector<pair<timeIter, timeIter> >::const_iterator cp = tempPulseStorage.begin(); cp < tempPulseStorage.end(); ++cp)
	  theCandidatePulses->push_back(*cp);
      }
  }


pair<float, float> PulseFinderProcessor::calculate_charge_and_time( const EVENT::TrackerData *pulse_data ) {

	// the values to return
	pair<float, float> charge_and_time;

	// check if the fit method is used
	if( _use_fit_charge || _use_fit_time ) {

		charge_and_time = calculate_charge_and_time_using_fit( pulse_data );
	}

	if( !_use_fit_charge ) {

		charge_and_time.first = integrate_pulse_charge( pulse_data );
	}

	if( !_use_fit_time ) {

		charge_and_time.second = calcPulseTime( pulse_data );
	}

	return charge_and_time;
}


pair<float, float> PulseFinderProcessor::calculate_charge_and_time_using_fit( const EVENT::TrackerData *pulse_data ) {

	// find the polarity of the pulse data
	const int polarity = get_polarity( pulse_data->getCellID1() );

	const vector<float> &charge_values = pulse_data->getChargeValues();

	// fit a parabola to the log of the data around the maximum

	// find the maximum (or minimum)
	vector<float>::const_iterator extreme_point = (polarity == -1.)? min_element( charge_values.begin(), charge_values.end() ): max_element( charge_values.begin(), charge_values.end() );

	// use 3 points on each side (at most)
	vector<float>::const_iterator fit_start = extreme_point;
	advance( fit_start, max( -3, (int)distance( fit_start, charge_values.begin() ) ) );

	vector<float>::const_iterator fit_end = extreme_point;
	advance( fit_end, min( 3, (int)distance( fit_end, charge_values.end() ) ) );

	// initialize the fit
	const int point_count = distance( fit_start, fit_end );
	gsl_matrix *X = gsl_matrix_alloc( point_count, 3 );
	gsl_vector *y = gsl_vector_alloc( point_count );
	gsl_vector *c = gsl_vector_alloc( 3 );
	gsl_matrix *covariance = gsl_matrix_alloc( 3, 3 );
	double minimum = 0.;

	streamlog_out( DEBUG0 ) << "Calculating pulse time using the points ";

	vector<float>::const_iterator current_charge = fit_start;

	for( int row = 0; current_charge != fit_end; current_charge++, row++ ) {

		// prevent bad log values
		if( (*current_charge * polarity) > 0 ) {

			gsl_vector_set( y, row, std::log(*current_charge * polarity) );
		}
		else {

			continue;
		}

		// may as well transform the fit in the t-direction
		gsl_matrix_set( X, row, 0, 1. );
		gsl_matrix_set( X, row, 1, row );
		gsl_matrix_set( X, row, 2, row*row );

		streamlog_out( DEBUG0 ) << '(' << row << ", " << *current_charge << ") ";
	}

	gsl_multifit_linear_workspace *workspace = gsl_multifit_linear_alloc( point_count, 3 );

	// perform the fit
	gsl_multifit_linear( X, y, c, covariance, &minimum, workspace );

	streamlog_out( DEBUG0 ) << ". Got: " << gsl_vector_get(c, 2) << "x^2 + " << gsl_vector_get(c, 1) << "x + " << gsl_vector_get(c, 0) << endl;

	// calculate the time
	// convert the result to nanoseconds
	const double translated_bin = -0.5 * gsl_vector_get( c, 1 ) / gsl_vector_get( c, 2 );
	const double bin_time = pulse_data->getTime() + translated_bin + distance( charge_values.begin(), fit_start ) + 0.5;

	const double time_bin_width = 1.E03 / get_readout_frequency( pulse_data->getCellID1() );
	const double fit_time = bin_time * time_bin_width;

	streamlog_out( DEBUG0 ) << "Time = " << translated_bin << " + 0.5 + " << distance( charge_values.begin(), fit_start ) << " + " << pulse_data->getTime() << " = " << bin_time << "bins" << endl;
	streamlog_out( DEBUG0 ) << "Bin width = " << time_bin_width << endl;

	// calculate the amplitude
	const float max_log = gsl_vector_get(c, 0) - ( 0.25 * gsl_vector_get(c, 1) * gsl_vector_get(c, 1) / gsl_vector_get(c, 2) );
	const float fit_amplitude = exp( max_log );

	// clean up the fit
	gsl_multifit_linear_free( workspace );
	gsl_matrix_free( X );
	gsl_vector_free( y );
	gsl_vector_free( c );

	return pair<int, int>( fit_amplitude, fit_time );
}


  float PulseFinderProcessor::calcPulseTime( const TrackerData* thisPulse )
  {
    vector<float> chargeVec = thisPulse->getChargeValues();

	// determine the polarity of the pulse from the electronics parameters
	const int polarity = get_polarity( thisPulse->getCellID1() );

	// calculate the noise
	float noise_value = get_pedestal_width( thisPulse->getCellID0(), thisPulse->getCellID1() ) * _pulseStartThreshold;

	// if pedestal width is not defined its default value is -1
	// then set pedestal width equal to 1
	if (noise_value < 0.)
	      noise_value = _pulseStartThreshold;
    
     float PulseTime    = 0.0;
     float Derivative   = 0.0;
     float Norm         = 0.0;
     float SliceTime    = thisPulse->getTime();

     if (chargeVec.size()>1)
       {
	 for(vector<float>::const_iterator timeBin = chargeVec.begin() + 1; timeBin < chargeVec.end(); ++timeBin)
	   { 
	     Derivative  = (*timeBin)-(*(timeBin-1));

	     // make derivative sign insensitive
		Derivative *= polarity;
	     
	     SliceTime += 1.;

	     // break out of calculation if decrease is larger than the noise
	     if (Derivative < (-noise_value))
	       break;

	     Norm      += Derivative;
	     PulseTime += Derivative * SliceTime;  
	   }
	 
	 PulseTime /= Norm;
       }
     else
       {
	 PulseTime = SliceTime;
       }

	const float bin_width = 1.E03 / get_readout_frequency( thisPulse->getCellID1() );
     
     return PulseTime*bin_width; //conversion in ns
  }
  
  
  
float PulseFinderProcessor::integrate_pulse_charge( const TrackerData* thisPulse ) {

	const int polarity = get_polarity( thisPulse->getCellID1() );
	const FloatVec charge_values = thisPulse->getChargeValues();
	
    float         PulseCharge  = 0.0;

    for (vector<float>::const_iterator adcVal = charge_values.begin(); adcVal < charge_values.end(); ++adcVal)
      PulseCharge += (*adcVal);
      
    /// flip sign if polarity is negative
	PulseCharge *= polarity;
    
    return PulseCharge;
}


/** Returns the pedestal width for a given TrackerData / TrackerPulse object taking into consideration whether it should be overwritten by a processor parameter
 */
float PulseFinderProcessor::get_pedestal_width( const int hardware_channel, const int readout_group ) {

	if( parameterSet( "PedestalWidthOverride") ) {

		return _pedestal_width_override;
	}
	else {

		#ifdef USE_LCCD
		return _pedestal_map->find( Pedestal::key_type( hardware_channel, readout_group ) ).getPedestalWidth();
		#else
		return _pedestalHandler->getPedestalWidth( hardware_channel, readout_group );
		#endif
	}
}


/** Returns the polarity of a TrackerData / TrackerPulse object taking into consideration whether it should be overwritten by a processor parameter
 */
int PulseFinderProcessor::get_polarity( const int readout_group ) {

	if( parameterSet("SignalPolarityOverride") ) {

		return _polarity_override;
	}
	else {

		return _electronic_parameters_map->find( GenericADCElectronicsParameters::key_type( readout_group ) ).getPolarity();
	}
}

/** Returns the readout frequency of a readout group (in Mhz) taking into consideration whether it should be overwritten by a processor parameter
 */
float PulseFinderProcessor::get_readout_frequency( const int readout_group ) {

	if( parameterSet("ReadoutFrequencyOverride") ) {

		return 1.E-06 * _readout_frequency_override;
	}
	else {

		// determine the readout frequency from LCCD (in MHz)
		return _electronic_parameters_map->find( readout_group ).getFrequency();
	}
}


/** Returns the maximum ADC value of a readout group taking into consideration whether it should be overwritten by a processor parameter
 */
float PulseFinderProcessor::get_max_ADC_value( const int readout_group ) {

	if( parameterSet("MaximumADCValueOverride") ) {

		return _maximumADCValueOverride;
	}
	else {

		return _electronic_parameters_map->find( readout_group ).getMaxADCValue();
	}
}


}
