#ifdef USE_TRACK_FITTER_KALMAN
#include "TrackMakingKalmanFilterProcessor.h"

#include <cmath>
#include <map>
#include <list>

#include <EVENT/LCCollection.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackImpl.h>
#include <IMPL/LCFlagImpl.h>
#include <marlin/Global.h>

#include <gearimpl/RectangularPadRowLayout.h>
#include <gear/PadRowLayout2D.h>
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>
#include <gearimpl/TPCParametersImpl.h>

#include "AttKalTrack.h"
#include "HeliKalTrack.h"
#include "StraightKalTrack.h"


using namespace marlin ;

namespace marlintpc
{
AttKalTrack * TrackMakingKalmanFilterProcessor::_theKalTrackPtr=0;

TrackMakingKalmanFilterProcessor aTrackMakingKalmanFilterProcessor;

TrackMakingKalmanFilterProcessor::TrackMakingKalmanFilterProcessor() : Processor("TrackMakingKalmanFilterProcessor")
{
   _description = "For simple track finding and fitting by Kalman filter!";
  
  registerInputCollection(   LCIO::TRACKERHIT,
			      "InputTrackerHits" , 
			      "Name of the Input TrackerHits collection"  ,
			      _InputTrackerHitsCollectionName ,
			      std::string("TPCHits") ) ;

  registerOutputCollection(   LCIO::TRACK,
			      "OutputTracks" , 
			      "Name of the output Tracks collection"  ,
			      _outputTracksCollectionName ,
			      std::string("TPCTracks") ) ;

  registerProcessorParameter( "MinTrackHits" , 
			      "Minimum number of hits on track (default: 15)",
			      _minTrackHits ,
			      int(15) ) ;

  registerProcessorParameter( "MaxSkipRows" , 
			      "Maximum number of subsequnetly missing hits (default: 5)",
			      _maxSkipRows ,
			      int(5) ) ;     
  registerProcessorParameter( "TrackType" ,
                              "Track type (default: 1, 0=straight line 1=helix)",
                              _trackType ,
                              int(1) ) ;
}


TrackMakingKalmanFilterProcessor::~TrackMakingKalmanFilterProcessor() 
{
}


void TrackMakingKalmanFilterProcessor::processRunHeader(LCRunHeader *run) 
{
    run->parameters().setValue( _processorName + "_revision", "$Rev: 2113 $");

    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
        if ( ! i->second->isOptional() || i->second->valueSet() ) {
            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }
}


void TrackMakingKalmanFilterProcessor::init() 
{
     if( streamlog::out.write< streamlog::MESSAGE3 >() ) printParameters();

     _tpcParameters = & Global::GEAR->getTPCParameters();

     //FIXME
     _moduleOrder=new map<int,int>;     
     const int NMODULE=7;// 7 modules
     for(int n=0;n<NMODULE;n++) _moduleOrder->insert(pair<int,int>(n,n));

     _hitMap=new TPCHitMap(_tpcParameters, _InputTrackerHitsCollectionName, _moduleOrder);

     if(_trackType==0)
     {
        _trackTuple=AIDAProcessor::tupleFactory(this)->create("tracktuple", 
                                                              "tracktuple", 
                                                              "double slope, double intercept");
        _theKalTrackPtr = new StraightKalTrack(_tpcParameters);
     }
     else
     { 
        _trackTuple=AIDAProcessor::tupleFactory(this)->create("tracktuple", 
                                                              "tracktuple", 
                                                              "double kappa, int ndf, double chi2");
        _theKalTrackPtr = new HeliKalTrack(_tpcParameters);
     }
    
     _theKalTrackPtr->setHitMap(_hitMap);
     _theKalTrackPtr->setMinTrackHits(_minTrackHits);
     _theKalTrackPtr->setMaxSkipRows(_maxSkipRows);
}


void TrackMakingKalmanFilterProcessor::processEvent(LCEvent *evt) 
{
    streamlog_out(MESSAGE4) << endl <<    "--------------------------------------"   << endl
                            << "Searching tracks in event " 
                            << evt->getEventNumber() << ":" << endl;

    _tpcTracksCollection = new LCCollectionVec(LCIO::TRACK);//deletion is left to processor

    //fill map with hits of this event
    _hitMap->fillEventHits(evt);
    
    //_theKalTrackPtr->drawEvent(evt, _InputTrackerHitsCollectionName, 1);

    //if there is hit in map, search track starting from a hit.
    //if a track is found, save the track.
    while(!_hitMap->isEmpty())
    {
       TrackImpl* aTrack;
       aTrack =_theKalTrackPtr->findTrack(_hitMap->getFirstHit());
       if(aTrack==0)continue;

       _theKalTrackPtr->printStateVector();

       if(_trackType==0)
       {
          StraightKalTrack* straightTrack=dynamic_cast<StraightKalTrack*> (_theKalTrackPtr);
	  double slope=straightTrack->getSlope();
          double intercept=straightTrack->getIntercept();

          _trackTuple -> fill(0, slope);
          _trackTuple -> fill(1, intercept);
          _trackTuple -> addRow();

       }
       else
       {
          HeliKalTrack* heliTrack=dynamic_cast<HeliKalTrack*>(_theKalTrackPtr);
          double kappa=heliTrack->getKappa();
          int    ndf=  heliTrack->getNDF();
          double chi2= heliTrack->getChi2();

          _trackTuple -> fill(0, kappa);
          _trackTuple -> fill(1, ndf);
          _trackTuple -> fill(2, chi2);
          _trackTuple -> addRow();
       }

       if(aTrack!=0)_tpcTracksCollection->addElement(aTrack);
    }

   //save track collection
   int trackNum = _tpcTracksCollection->getNumberOfElements();
   if(trackNum!=0)
   {
       evt->addCollection( _tpcTracksCollection,_outputTracksCollectionName );
       streamlog_out(MESSAGE4) << "Event " << evt->getEventNumber() 
                               << "  has " << trackNum << " tracks" 
                               << endl;
   }
}


Processor* TrackMakingKalmanFilterProcessor::newProcessor() 
{
   return new TrackMakingKalmanFilterProcessor();
}


void TrackMakingKalmanFilterProcessor::check(LCEvent *evt) 
{
}


void TrackMakingKalmanFilterProcessor::end() 
{
   delete _theKalTrackPtr;
   delete _hitMap;
   delete _trackTuple;

   streamlog_out(MESSAGE4) << "End of TrackMakingKalmanFilterProcessor" << endl;
}

} // namespace marlintpc
#endif
