#include "TrackerRawDataToDataConverterProcessor.h"

// C++
#include <iostream>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>

// Aida
#ifdef MARLIN_USE_AIDA
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/ICloud1D.h>
#endif

using namespace lcio;
using namespace marlin;
using namespace std;

namespace marlintpc
{
  TrackerRawDataToDataConverterProcessor aTrackerRawDataToDataConverterProcessor;
  
  TrackerRawDataToDataConverterProcessor::TrackerRawDataToDataConverterProcessor() : Processor("TrackerRawDataToDataConverterProcessor") 
  {
    /* the constructor call first the super constructor which the name 
     * of the processor
     * this must by the same as the class name.
     */
    
    /* description of the processor which will be displayed in the
     * steering file autometicly generated by 
     * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
     */
    _description = "TrackerRawDataToDataConverterProcessor converts the TrackerRawData to TrackerData without processing the values";
    
    /* register steering parameters: name, description, class-variable, default value
     * the type will by definded by the type of the default value
     * string, double, float and int are possible
     * please note to get a float you must cast in C++: (float) 1.3 
     */
    registerInputCollection(LCIO::TRACKERRAWDATA,
			    "InputCollectionName" , 
			    "Name of the input collection"  ,
			    _inputColName ,
			    std::string("TPCRawData") );
    
    registerOutputCollection(LCIO::TRACKERDATA,
			     "OutputCollectionName" , 
			     "Name of the output collection"  ,
			     _outputColName ,
			     std::string("TPCConvertedRawData") );
    
    registerProcessorParameter("SetOutputTransient" , 
			       "if not 0 the output collection is set transient"  ,
			       _outputIsTransient,
			       true);
  }
  
  
  
  void TrackerRawDataToDataConverterProcessor::init()
  { 
    printParameters();
    
    _nRun = 0;
    _nEvt = 0;
  }
  
  
  
  void TrackerRawDataToDataConverterProcessor::processRunHeader(LCRunHeader* run)
  { 
    run->parameters().setValue( _processorName + "_revision", "$Rev: 2111 $");
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) 
      {
	if ( ! i->second->isOptional() || i->second->valueSet() ) 
	  {
	    run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
	  }
      }
    
    _nRun++;
  } 
  


  void TrackerRawDataToDataConverterProcessor::processEvent(LCEvent* evt)
  { 
    // get the input collection from the event
    try
      {
	LCCollection* inputCol = evt->getCollection( _inputColName );

	m_out(MESSAGE) << "Processing " << inputCol->getNumberOfElements() << " TrackerRawData elements from collection " << _inputColName << "." << endl;
	
	// create the output collection with a string defining the TYPE (see lcio.h)
	LCCollectionVec* outputCol = new LCCollectionVec( LCIO::TRACKERDATA ); 
	
	// set the transient flag as definded by the processor parameter
	outputCol->setTransient( (_outputIsTransient != 0) );
	
	// get the number of elements in the input collection and loop over it
	for(int iRawData=0; iRawData< inputCol->getNumberOfElements(); iRawData++)
	  {
	    /* getElement return only a LCObject
	     * you have to cast it the get the right type
	     * please use dynamic_cast which will return 0 if the object can not be casted
	     */
	    TrackerRawData* RawData = dynamic_cast<TrackerRawData*>( inputCol->getElementAt( iRawData ) );
	    
	    if ( RawData != 0 ) 
	      {
		//only poceed if the cast was successful
		
		/* create a new object of the TrackerData type
		 * you have to use the IMPL, otherwise you can not set values
		 */
		TrackerDataImpl* Data = new TrackerDataImpl();
		
		// transfer the values
		
		//first the simple
		Data->setCellID0( RawData->getCellID0() );
		Data->setCellID1( RawData->getCellID1() );// normally 0 for TPC data 
		
		Data->setTime( (float)RawData->getTime() );
		
		// and now the vector
		ShortVec RawDataADCVal = RawData->getADCValues();
		
		for (ShortVec::iterator iSlice=RawDataADCVal.begin(); iSlice!=RawDataADCVal.end(); iSlice++)
		  {
		    // fill the data vector with 
		    (Data->chargeValues()).push_back( (float)(*iSlice) );
		  }
		
		/* fill the TrackerData in the collection
		 * please note: DO NOT DELETE the object this will be done by LCIO
		 */
		outputCol->addElement( Data );
	      }
	  }
	
	/* after converting all raw data objects
	 * add the collection to the eventand asigning a name to it
	 * please note again: DO NOT DELETE the collection this will be done by LCIO
	 */
	evt->addCollection( outputCol, _outputColName );
	
      }
    catch(DataNotAvailableException)
      {
	// if there is not input collection give a warning to the user
	// note: this is not a graceful way to handle this exception
	m_out(WARNING) << "Warning of TrackerRawDataToDataConverterProcessor::processEvent: input collection " 
		       << _inputColName << " cannot be found in event " << _nEvt << m_endl;	
      }
    
    _nEvt++;
  }


  
  void TrackerRawDataToDataConverterProcessor::check(LCEvent* evt)
  { 
    // nothing to check here - could be used to fill checkplots in reconstruction processor
  }


  
  void TrackerRawDataToDataConverterProcessor::end()
  { 
    //   nothing done here
  }
}
