
// LCIO
#include <EVENT/LCIO.h>
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/LCFlagImpl.h>

// stl
#include <map>
#include <cassert>

// GEAR
#include <gear/GEAR.h>
#include <gear/TPCParameters.h>

// Marlin
#include <marlin/Global.h>
#include <marlin/Exceptions.h>

// MarlinTPC
#include "SignalCombinerProcessor.h"

using namespace std;
using namespace EVENT;
using namespace IMPL;
using namespace marlin;

namespace marlintpc {

// create one to register it with Marlin
SignalCombinerProcessor aSignalCombinerProcessor;


SignalCombinerProcessor::SignalCombinerProcessor():
	Processor("SignalCombinerProcessor"),
	_input_collection_name("TPCShapedSignals"),
	_output_collection_name("TPCTrackerData"),
	_time_resolution(50.) {

	registerInputCollection(
		LCIO::LCGENERICOBJECT,
		"InputTPCShapedSignals",
		"Name of the input TrackerData collection containing individually shaped pulses",
		_input_collection_name,
		string("TPCShapedSignals") );

	registerOutputCollection(
		LCIO::TRACKERDATA,
		"OutputTPCTrackerData",
		"Name of the output TrackerData collection.",
		_output_collection_name,
		string("TPCTrackerData") );

	registerOptionalParameter(
		"TimeResolution",
		"The width of a signal sample (1/Hz) in ns.",
		_time_resolution,
		50.);
}


SignalCombinerProcessor::~SignalCombinerProcessor() {}


void SignalCombinerProcessor::processRunHeader( LCRunHeader* run) {

	// store the processor parameters in the event
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {

        if ( ! i->second->isOptional() || i->second->valueSet() ) {

            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }
}


void SignalCombinerProcessor::processEvent( LCEvent * evt ) {

	// grab the collection of energy deposits from the simulation
	LCCollection *individual_trackerdata_collection = NULL;
	try {

		individual_trackerdata_collection = evt->getCollection( _input_collection_name );
	}
	catch (DataNotAvailableException &e) {

		// the collection of charge signals with that name wasn't in the LCIO event file, skip simulating this event
		throw SkipEventException( this );
	}

	streamlog_out( MESSAGE0 ) << "Combining " << individual_trackerdata_collection->getNumberOfElements() << " signals\n";

	// get the width of a time bin in ns
	if( !parameterSet("TimeResolution") ) {

		_time_resolution = 1.E09 / Global::GEAR->getTPCParameters().getReadoutFrequency();
	}

	// create a collection to store the combined tracker data
	LCCollectionVec *trackerdata_collection = new LCCollectionVec( LCIO::TRACKERDATA );
	trackerdata_collection->setTransient(true);
	trackerdata_collection->setSubset(false);

	// set the flags that cellID1 should be stored
	lcio::LCFlagImpl trkFlag(0) ;
	trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
	trackerdata_collection->setFlag( trkFlag.getFlag() );

	// a map containing the signals on each pad
	multimap<DataTag, TrackerData*> _data_by_tag;

	// now loop through single pulses mapping them by pad
	for( int trackerdata_number = 0; trackerdata_number < individual_trackerdata_collection->getNumberOfElements(); trackerdata_number++ ) {

		TrackerData *current_data = dynamic_cast<TrackerData*>( individual_trackerdata_collection->getElementAt( trackerdata_number ) );
		assert(current_data);

		DataTag tag( current_data->getCellID0(), current_data->getCellID1() );

		_data_by_tag.insert( pair<DataTag, TrackerData*>(tag, current_data) );
	}


	DataTag last_tag(-1, -1);
	TrackerDataImpl *combined_data;

	// now loop through the map combining the data for each pad
	for ( multimap<DataTag, TrackerData*>::iterator tag_data = _data_by_tag.begin() ; tag_data != _data_by_tag.end(); tag_data++ ) {

		if( last_tag != (*tag_data).first ) {

			combined_data = new TrackerDataImpl();
			combined_data->setCellID0( (*tag_data).first.pad_id );
			combined_data->setCellID1( (*tag_data).first.module_id );
			last_tag = (*tag_data).first;
			trackerdata_collection->addElement( combined_data );
		}

		// and combine them together
		combine_data( *(*tag_data).second, *combined_data, _time_resolution );
	}

	streamlog_out( MESSAGE0 ) << "Combined " << individual_trackerdata_collection->getNumberOfElements() << " individual signals into " << trackerdata_collection->getNumberOfElements() << " signals\n";

	// now add the collection to the event so other processors may access it
	try {

		evt->addCollection( trackerdata_collection, _output_collection_name );
	}
	catch( ... ) {

		throw SkipEventException( this );
	}
	
}


void SignalCombinerProcessor::combine_data( const TrackerData &new_data, TrackerDataImpl &combined_data, const double time_resolution ) {

	const double old_signal_duration = combined_data.chargeValues().size() * time_resolution;

	// this is a new signal, just copy the data over
	if( old_signal_duration == 0 ) {

		combined_data.setTime( new_data.getTime() );
		combined_data.chargeValues().assign( new_data.getChargeValues().begin(), new_data.getChargeValues().end() );
		return;
	}

	FloatVec &data = combined_data.chargeValues();

	// add this signal to the current one
	const double new_signal_duration = combined_data.getChargeValues().size() * time_resolution;

	const double old_start_time = combined_data.getTime();
	const double new_start_time = new_data.getTime();

	double offset = new_start_time - old_start_time;

	if( offset < 0. ) {

		combined_data.setTime( new_start_time );

		const int padding = (int) (-1 * offset / time_resolution);

		data.insert( data.begin(), padding, 0. );
	}

	offset = new_start_time + new_signal_duration - (old_start_time + old_signal_duration);

	while( offset > 0 ) {

		data.push_back(0.);
		offset -= time_resolution;
	}

	// now that there is room for the new signal, add it
	int current_bin = (int) ( (new_start_time - combined_data.getTime() ) / time_resolution );

	for( FloatVec::const_iterator new_bin = new_data.getChargeValues().begin(); new_bin != new_data.getChargeValues().end(); new_bin++ ) {

		data[current_bin] += (*new_bin);
		current_bin++;
	}
}


void SignalCombinerProcessor::init() {

    // usually a good idea to
    printParameters() ;
}


void SignalCombinerProcessor::check( LCEvent * evt ) {

}


void SignalCombinerProcessor::end() {

}

}
