// MarlinTPC
#include "ElectronCloud.h"
#include "PhotoelectricPulseProcessor.h"

// Marlin
#include <marlin/Exceptions.h>

// LCIO
#include <EVENT/LCIO.h>
#include <EVENT/LCCollection.h>
#include <EVENT/MCParticle.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/SimTrackerHitImpl.h>

// GEAR
#include <gear/TPCModule.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>
#include <gearxml/tinyxml.h>

// CLHEP
#include "CLHEP/Random/RandPoisson.h"

// stl
#include <iostream>
#include <fstream>

using namespace std;
using namespace EVENT;
using namespace IMPL;
using namespace marlin;
using namespace CLHEP;
using namespace gear;

namespace marlintpc {

// create one to register it with Marlin
PhotoelectricPulseProcessor aPhotoelectricPulseProcessor;


PhotoelectricPulseProcessor::PhotoelectricPulseProcessor():
	Processor("PhotoelectricPulseProcessor"),
	_photogeometry_filename("photo_geometry.txt"),
	_output_collection_name ( "TPCElectronClouds" ),
	_output_simtrackerhit_collection_name( "TPCSimPhotodots" ),
	_mean_electrons_per_photodot(100.) {

	registerOutputCollection(
		LCIO::LCGENERICOBJECT,
		"OutputTPCElectronClouds",
		"Name of the output LCGeneric collection full of electron clouds",
		_output_collection_name,
		string("TPCElectronClouds") );

	registerOutputCollection(
		LCIO::SIMTRACKERHIT,
		"OutputSimPhotodots",
		"Name of the output SimTrackerHit collection containing the true centroids of the photodots.",
		_output_simtrackerhit_collection_name,
		string("TPCSimPhotodots") );

	registerOptionalParameter(
		"GeometryFilename",
		"The name of the file containing the positions of the photodots and photolines.",
		_photogeometry_filename,
		string("photo_geometry.txt") );

	registerOptionalParameter(
		"MeanElectronsPerPhotodot",
		"The mean number of electrons to create per photodot (Poisson distribution)",
		_mean_electrons_per_photodot,
		100.);

	registerOptionalParameter(
		"InitialZPosition",
		"The initial z position of the electron clouds.",
		_initial_z_position,
		0.);
}


PhotoelectricPulseProcessor::~PhotoelectricPulseProcessor() {}


void PhotoelectricPulseProcessor::processRunHeader( LCRunHeader* run) {

	// store the processor parameters in the event
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {

        if ( ! i->second->isOptional() || i->second->valueSet() ) {

            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }
}


void PhotoelectricPulseProcessor::processEvent( LCEvent * evt ) {

	// create a collection to store the electron clouds (don't write this collection at the end)
	LCCollectionVec *electroncloud_collection = new LCCollectionVec( LCIO::LCGENERICOBJECT );
	electroncloud_collection->setTransient();

	// create a collection to store the SimTrackerHits
	LCCollectionVec *simtrackerhit_collection = new LCCollectionVec( LCIO::SIMTRACKERHIT );
	electroncloud_collection->setTransient();

	// keep track of the number of clouds produced for output
	unsigned int number_of_electron_clouds = 0;

	// now loop through the photodots creating clouds of electrons
	for( unsigned int photodot_number = 0; photodot_number < _photodots.size(); photodot_number++ ) {

		Photodot &photodot = _photodots[photodot_number];

		const int number_of_electrons = RandPoisson::shoot(_mean_electrons_per_photodot);

		if( number_of_electrons <= 0 ) {

			continue;
		}

		// create an electron cloud
		ElectronCloud *new_cloud = new ElectronCloud();

		// set up the deviation in the position of the electrons in the cloud
		new_cloud->set_transverse_stdev( photodot._diameter/4. );
		new_cloud->set_longitudinal_stdev(0.01);

		new_cloud->set_cloud_id( photodot_number );

		// set up the mean position of the electrons in the cloud
		new_cloud->set_x_position( photodot._x );
		new_cloud->set_y_position( photodot._y );
		new_cloud->set_z_position( _initial_z_position );

		streamlog_out( DEBUG ) << "Produced a cloud with " << number_of_electrons << " electrons at \t" << photodot._x << "mm, \t" << photodot._y << "mm\n";

		// set up the other parameters
		new_cloud->set_time( 0. );
		new_cloud->set_electron_count( number_of_electrons );

		// now add the cloud to the collection
		electroncloud_collection->addElement( new_cloud );

		// create a SimTrackerHit which holds the true centroid of the photodot
		SimTrackerHitImpl *simtrackerhit = new SimTrackerHitImpl();

		double position[3];
		position[0] = photodot._x;
		position[1] = photodot._y;
		position[2] = _initial_z_position;
		simtrackerhit->setPosition( position );

		simtrackerhit_collection->addElement( simtrackerhit );

		number_of_electron_clouds += 1;
	}

	streamlog_out( MESSAGE ) << "Produced " << number_of_electron_clouds << " electron clouds.\n";

	// now add the collection to the event so other processors may access it
	try {

		evt->addCollection( electroncloud_collection, _output_collection_name );
		evt->addCollection( simtrackerhit_collection, _output_simtrackerhit_collection_name );
	}
	catch( ... ) {

		throw SkipEventException( this );
	}
	
}


void PhotoelectricPulseProcessor::init() {

	// usually a good idea to
	printParameters();

	// open the projection file and get the time interval we're watching for and the list of pads to take
	TiXmlDocument document ( _photogeometry_filename.c_str() );

	if( !document.LoadFile() ) {

		streamlog_out( ERROR ) << "Couldn't load projected geometry file " << _photogeometry_filename << "." << endl;
		return;
	}

	// get a list of the input photoelectric dots
	int i_dot = 0;
	TiXmlElement *photodot_element = document.RootElement()->FirstChildElement("photodot");

	for ( ; photodot_element; photodot_element = photodot_element->NextSiblingElement( "photodot" ) ) {

		// get the diameter
		double diameter;
		photodot_element->QueryDoubleAttribute ( "diameter", &diameter );

		// and the centroid
		TiXmlElement *centroid_element = photodot_element->FirstChildElement( "centroid" );

		if ( centroid_element ) {

			string centroid_text = centroid_element->FirstChild()->Value();

			size_t comma_pos = centroid_text.find (',');

			string coord_text = centroid_text.substr ( 1, comma_pos - 1 );
			const float x = atof ( coord_text.c_str() );

			coord_text = centroid_text.substr ( comma_pos + 1, centroid_text.length() - comma_pos - 1 );
			const float y = atof ( coord_text.c_str() );

			_photodots.push_back( Photodot( x, y, diameter ) );

			streamlog_out( DEBUG ) << "photodot " << i_dot << " has centroid " << centroid_text << " and diameter " << diameter << "." << endl;
		}

		i_dot++;
	}

	// get a list of the input photoelectric lines
	int i_line = 0;
	TiXmlElement *photoline_element = document.RootElement()->FirstChildElement("photoline");

	for ( ; photoline_element; photoline_element = photoline_element->NextSiblingElement( "photoline" ) ) {

		// get the width
		double width;
		photoline_element->QueryDoubleAttribute ( "width", &width );

		// and the endpoints
		double x1, y1, x2, y2;
		TiXmlElement *point_element = photoline_element->FirstChildElement( "start_point" );

		if ( point_element ) {

			string point_text = point_element->FirstChild()->Value();

			size_t comma_pos = point_text.find (',');

			string coord_text = point_text.substr ( 1, comma_pos - 1 );
			x1 = atof ( coord_text.c_str() );

			coord_text = point_text.substr ( comma_pos + 1, point_text.length() - comma_pos - 1 );
			y1 = atof ( coord_text.c_str() );
		}
		else {

			streamlog_out( WARNING ) << "photoline " << i_line << " is missing the start_point element." << endl;
			continue;
		}

		point_element = photoline_element->FirstChildElement( "end_point" );

		if ( point_element ) {

			string point_text = point_element->FirstChild()->Value();

			size_t comma_pos = point_text.find (',');

			string coord_text = point_text.substr ( 1, comma_pos - 1 );
			x2 = atof ( coord_text.c_str() );

			coord_text = point_text.substr ( comma_pos + 1, point_text.length() - comma_pos - 1 );
			y2 = atof ( coord_text.c_str() );
		}
		else {

			streamlog_out( WARNING ) << "photoline " << i_line << " is missing the end_point element." << endl;
			continue;
		}

		_photolines.push_back( Photoline( x1, y1, x2, y2, width ) );

		streamlog_out( DEBUG ) << "photodot " << i_line << " has start point (" << x1 << ", " << y1 << ") end point (" << x2 << ", " << y2 << ") and width " << width << "." << endl;

		i_line++;
	}
}


void PhotoelectricPulseProcessor::check( LCEvent * evt ) { 

}


void PhotoelectricPulseProcessor::end() { 

}

}
