
// stl
#include <string>
#include <cassert>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/LCGenericObject.h>

// marlintpc
#include "SolenoidZMapMagneticField.h"

using namespace CLHEP;
using namespace EVENT;
using namespace std;

namespace marlintpc {

SolenoidZMapMagneticField aSolenoidZMapMagneticField;

/** constructor
 */
SolenoidZMapMagneticField::SolenoidZMapMagneticField():
	Field("SolenoidZMapMagneticField") {
}


/** destructor
 */
SolenoidZMapMagneticField::~SolenoidZMapMagneticField() {}


/** returns a new field... similar to how the Marlin processor registration works
*/
Field *SolenoidZMapMagneticField::new_field( void ) { return new SolenoidZMapMagneticField; }


/** constructs the field from an LCCollection of data objects (could be anything... field bin values or just parameters)
*/
bool SolenoidZMapMagneticField::construct( LCCollection *collection ) {

	_field_map.clear();

	// grab the bounds from the collection
	_r_min = collection->getParameters().getFloatVal("r_min");
	_r_max = collection->getParameters().getFloatVal("r_max");
	_dz = collection->getParameters().getFloatVal("dz");

	for( int map_counter = 0; map_counter < collection->getNumberOfElements(); map_counter++ ) {

		LCGenericObject *field_map_object = dynamic_cast<LCGenericObject*>( collection->getElementAt(map_counter) );
		assert(field_map_object);

		FieldMapEntry map_entry;

		map_entry._z = field_map_object->getDoubleVal(0);
		map_entry._Bz = field_map_object->getDoubleVal(1);

		_field_map.push_back(map_entry);
	}

	return true;
}


/** gets the field value at a specified point
*/
CLHEP::Hep3Vector SolenoidZMapMagneticField::get_field_value( const CLHEP::Hep3Vector &position ) {

	const int signZ = (position.z() >= 0) ? 1: -1; // neglect the case z = 0 here...

	const double	raw_index	= fabs(position.z() / _dz); // the exact position in units of bins (fractional)
	const unsigned int  floor_index = static_cast<unsigned int>(raw_index + 0.0); // the lower neighbouring bin (integer)
	const unsigned int  round_index = static_cast<unsigned int>(raw_index + 0.5); // the closest of the two neighbouring bins (integer)
	const unsigned int  ceil_index 	= static_cast<unsigned int>(raw_index + 1.0); // the higher neighbouring bin (integer)

	CLHEP::Hep3Vector field_vector(0., 0., 0.);

	if (ceil_index >= _field_map.size()) return field_vector;

	const double floor_weight = ceil_index - raw_index; // how close are we to the lower bin? (one minus distance)
	const double ceil_weight = raw_index - floor_index; // how close are we to the higher bin? (one minus distance)
	const double Bz = floor_weight * _field_map[floor_index]._Bz + ceil_weight * _field_map[ceil_index]._Bz; // interpolation

	field_vector.setZ( Bz );
	
	// interpolation of the differential quotient needs two slopes, i.e. three points

	const int prev_index = (round_index == 0)            ? (round_index) : (round_index - 1); // the previous bin
	const int next_index = (round_index == _field_map.size() - 1) ? (round_index) : (round_index + 1); // the next bin
	
	const double prev_weight = (round_index - raw_index) + 0.5; // how close are we to the previous slope?
	const double next_weight = (raw_index - round_index) + 0.5; // how close are we to the next slope?
	
	const double prev_dBdz = (_field_map[round_index]._Bz - _field_map[prev_index]._Bz) / _dz; // the previous slope
	const double next_dBdz = (_field_map[next_index]._Bz - _field_map[round_index]._Bz) / _dz; // the next slope
	const double dBdz = prev_weight * prev_dBdz + next_weight * next_dBdz; // interpolation of slopes
	
	field_vector.setX( -0.5 * position.x() * dBdz * signZ );
	field_vector.setY( -0.5 * position.y() * dBdz * signZ );

	return field_vector;
}

} // namespace marlintpc
