#ifndef TimePixEventHistoProcessor_h
#define TimePixEventHistoProcessor_h 1

// C++
#include <string>
#include <map>

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "TimePixMapListener.h"
#include "TimePixPixelMode.h"
#include <marlin/AIDAProcessor.h>

namespace AIDA{
    class IHistogram2D;  
}

namespace marlintpc{

    /* the following comment is tranfered to doxygen documentation
     * which begins with a second asterix, or in case of on line comment an extra third slash 
     */
    /**  Creates one aida histogram per chip and event.
     *	 Can use zero suppressed or non zero suppressed TimePixData
     * 
     *
     *  <h4>Input - Prerequisites</h4>
     *  \li A collection of TrackerData
     *  \li A mode map for each chip
     *
     *  <h4>Output</h4> 
     *  \li One "event display" histogram per chip and event
     *
     *  @param InputCollectionName Name of the input collection (default TimePixZeroSuppressedData)
     *  @param HistoBaseName Base name of the histograms (default TimePixEvent)
     *  @param nColumn Number of columns of the chip (default 256)
     *  @param nRow Number of rows of the chip (default 256)
     *  @param Mode Mode that should be plotted (TOT, TIMEPIX, MEDIPIX, ONEHIT or ANY), default: ANY
     * 
     * @author Simone Zimmermann, Frederick Klöckner, Martin Killenberg
     */
    class TimePixEventHistoProcessor : public marlin::Processor {
	
    public:
	
	/* This method will be called by the marlin package
	 * It returns a processor of the currend type (here TimePixEventHistoProcessor)
	 */
	virtual Processor*  newProcessor() { return new TimePixEventHistoProcessor ; }
	
	/* the default constructor
	 * here the processor parameters are registered to the marlin package
	 * other initialisation should be placed in the init method
	 */
	TimePixEventHistoProcessor() ;
	
	/* Called at the beginning of the job before anything is read.
	 * Use to initialize the processor, e.g. book histograms
	 */
	virtual void init() ;
	
	/* Called for every run.
	 * in this processor it is not used
	 */
	virtual void processRunHeader( lcio::LCRunHeader* run ) ;
	
	/* Called for every event - the working horse.
	 * Here the real conversion take place
	 */
	virtual void processEvent( lcio::LCEvent * evt ) ; 
	
	/* This method is only called if the check flag is set (default) in the main processor
	 * (where the input files are given)
	 * It should be used to create check plots which are not needed in a complete analysis
	 * (where normaly the check flag is deactivated)
	 */
	virtual void check( lcio::LCEvent * evt ) ; 
	
	
	/* Called after data processing for clean up.
	 * e.g. saving control histogramms, delete created objects
	 * (which are not stored in a collection/event), etc.
	 */
	virtual void end() ;
	
	
    protected:
	
	/* normally the doxygen documentation is palced before the method/variable
	 * but for one line comments in case of variables it may be more convenied for you
	 * to place it after the variable. In this case you use an arrow to tell doxygen
	 * to which variable the commet belongs to.
	 */
	int _nColumn; ///< number of columns of the chip
	int _nRow; ///< number of rows of the chip
	std::string _mode; ///< mode that should be plotted
	std::string _inputColName; ///< Name of the input collection
	std::string _histoBaseName; ///< Base name of the histograms

	/** A helper class to store AIDA histograms in a map.
	 *  This requires an object with a propper constructor, which AIDA
	 *  does not provide because you have to use the factory.
	 */ 
	class AIDAHistogram
	{
	   public:
	    AIDAHistogram();
	    ~AIDAHistogram();
	    AIDA::IHistogram2D *histo;
	};

	/// A map with the chipID as key and an AIDAHistogram as value
	std::map<int, AIDAHistogram> _eventHistos;
	tpcconddata::TimePixMap * _timePixMap;

    } ;
    
}

#endif
    
    

